
import sqlite3, logging
READBUF = 8192

class MD5_Cache:

    def __init__(self, cache_db="/tmp/#vos_cached.db#"):
        """Setup the sqlDB that will contain the cache table"""
        self.cache_db = cache_db

        ## initialize the md5Cache db
        sqlConn = sqlite3.connect(self.cache_db)
        with sqlConn:
            sqlConn.execute("create table if not exists md5_cache (fname text PRIMARY KEY NOT NULL , md5 text, st_size int, st_mtime int)")
        ## build cache lookup if doesn't already exists

    def computeMD5(self, filename, block_size=READBUF):
        import hashlib
        md5 = hashlib.md5()
        r = open(filename, 'r')
        while True:
            buf = r.read(block_size)
            if len(buf) == 0:
                break
            md5.update(buf)
        r.close()
        return md5.hexdigest()


    def get(self, fname):
        """Get the MD5 for this fname from the SQL cache"""
        sqlConn = sqlite3.connect(self.cache_db)
        with sqlConn:
            cursor = sqlConn.execute("SELECT md5, st_size, st_mtime FROM md5_cache WHERE fname = ? ", (fname,))
            md5Row = cursor.fetchone()
        if md5Row is not None:
            return md5Row
        else:
            return None

    def delete(self, fname):
        """Delete a record from the cache MD5 database"""
        sqlConn = sqlite3.connect(self.cache_db)
        with sqlConn:
            sqlConn.execute("DELETE from md5_cache WHERE fname = ?", (fname,))


    def update(self, fname, md5, st_size, st_mtime):
        """Update a record in the cache MD5 database"""
        ## UPDATE the MD5 database
        sqlConn = sqlite3.connect(self.cache_db)
        try:
            with sqlConn:
                sqlConn.execute("DELETE from md5_cache WHERE fname = ?", (fname,))
                sqlConn.execute("INSERT INTO md5_cache (fname, md5, st_size, st_mtime) VALUES ( ?, ?, ?, ?)", (fname, md5, st_size, st_mtime))
        except Exception as e:
            logging.error(e)
        return md5




#!/usr/bin/env python2.7
"""A FUSE based filesystem view of VOSpace."""
import time
import vos
import sys
import urllib
from fuse import Operations, FuseOSError
from threading import Lock
from errno import EIO, ENOENT, EPERM, EAGAIN, EFAULT
import os
from os import O_RDONLY, O_WRONLY, O_RDWR, O_APPEND
from CadcCache import Cache, CacheCondition, CacheRetry, CacheAborted, \
    IOProxy, FlushNodeQueue, CacheError
from logExceptions import logExceptions
import logging

logger = logging.getLogger('vofs')
logger.setLevel(logging.ERROR)
if sys.version_info[1] > 6:
    logger.addHandler(logging.NullHandler())


def flag2mode(flags):
    md = {O_RDONLY: 'r', O_WRONLY: 'w', O_RDWR: 'w+'}
    m = md[flags & (O_RDONLY | O_WRONLY | O_RDWR)]

    # If the write bit was set, check to see if the append bit was also set.
    if flags | O_APPEND:
        m = m.replace('w', 'a', 1)

    return m


class MyIOProxy(IOProxy):

    def delNode(self, force=False):
        raise NotImplementedError("MyIOProxy.delNode")

    def __init__(self, vofs, path):
        super(MyIOProxy, self).__init__()
        self.vofs = vofs
        # This is the vos.VOFile object last used
        self.lastVOFile = None
        self.size = None
        self.md5 = None
        self.path = path
        self.condition = CacheCondition(None)

    def __str__(self):
        return "Path:{0}  Size:{1}  MD5:{2}  condition:{3}".format(self.path,
                                                                   self.size,
                                                                   self.md5,
                                                                   self.condition)

    @logExceptions()
    def writeToBacking(self):
        """
        Write a file in the cache to the remote file.
        """
        logger.debug("PUSHING %s to VOSpace @ %s" % (self.cacheFile.cacheDataFile, self.cacheFile.path))
        logger.debug("opening a new vo file for {0}".format(self.cacheFile.path))
        dest_uri = self.vofs.get_node(self.cacheFile.path).uri
        return self.vofs.client.copy(self.cacheFile.cacheDataFile, dest_uri, send_md5=True)

    @logExceptions()
    def readFromBacking(self, size=None, offset=0, block_size=Cache.IO_BLOCK_SIZE):
        """
        Read from VOSpace into cache
        """

        # Read a range
        if size is not None or offset != 0:
            if size is None:
                end_str = ""
            else:
                end_str = str(offset + size - 1)
            byte_range = "bytes=%s-%s" % (str(offset), end_str)
        else:
            byte_range = None
        logger.debug("reading range: %s" % (str(byte_range)))

        if self.lastVOFile is None:
            logger.debug("Opening a new vo file on {0}".format(self.cacheFile.path))
            self.lastVOFile = self.vofs.client.open(self.cacheFile.path,
                                                    mode=os.O_RDONLY, view="data", size=size, byte_range=byte_range,
                                                    possible_partial_read=True)
        else:
            logger.debug("Opening a existing vo file on {0}".format(self.lastVOFile.URLs[self.lastVOFile.urlIndex]))
            self.lastVOFile.open(
                self.lastVOFile.URLs[self.lastVOFile.urlIndex], byte_range=byte_range, possible_partial_read=True)
        try:
            logger.debug("reading from {0}".format(self.lastVOFile.URLs[self.lastVOFile.urlIndex]))
            try:
                resp = self.lastVOFile.read(return_response=True)
            except OSError as os_error:
                # existing URLs do not work anymore. Try another
                # transfer, forcing a full negotiation. This
                # handles the case that we tried a short cut URL
                # and it failed, so now we can try the full URL
                # list. If it still fails let the error propagate
                # to client
                self.lastVOFile = self.vofs.client.open(
                    self.cacheFile.path, mode=os.O_RDONLY, view="data",
                    size=size, byte_range=byte_range, full_negotiation=True, possible_partial_read=True)
                resp = self.lastVOFile.read(return_response=True)

            if not self.cacheFile.gotHeader:
                info = self.lastVOFile.get_file_info()
                logger.debug("Got info: {0}".format(info))
                self.cacheFile.setHeader(info[0], info[1])

            for buff in resp.iter_content(block_size):
                try:
                    self.writeToCache(buff, offset)
                except CacheAborted as ca:
                    logger.debug("Aborting write to cache: {0}".format(ca))
                    # The transfer was aborted.
                    break
                offset += len(buff)
        except Exception as e:
            self.exception = e
            raise e
        finally:
            self.lastVOFile.close()
            self.lastVOFile = None

        logger.debug("Wrote: %d bytes to cache for %s" % (offset, self.cacheFile.path))

    def get_md5(self):
        if self.md5 is None:
            node = self.vofs.get_node(self.path)
            self.set_size(int(node.props.get('length')))
            self.set_md5(node.props.get('MD5'))

        return self.md5

    def getSize(self):
        if self.size is None:
            node = self.vofs.get_node(self.path)
            self.set_size(int(node.props.get('length')))
            self.set_md5(node.props.get('MD5'))

        return self.size

    def set_md5(self, md5):
        """
        set the value of the MD5 sum returned for this location.

        @rtype : str
        @param md5: str representation of the MD5 sum of a file.
        """
        self.md5 = md5

    def set_size(self, size):
        self.size = size


class HandleWrapper(object):
    """
    Wrapper for cache file handlers. Each wrapper represents an open request.
    Multiple wrappers of the same file share the same cache file handler.
    """

    # A list of all file handles
    handle_list = {}
    handle_lock = Lock()

    def __init__(self, cache_file_handle, read_only=False):
        self.cache_file_handle = cache_file_handle
        self.read_only = read_only
        with HandleWrapper.handle_lock:
            HandleWrapper.handle_list[id(self)] = self

    def get_id(self):
        return id(self)

    @staticmethod
    def file_handle(file_id):
        with HandleWrapper.handle_lock:
            the_handle = HandleWrapper.handle_list[file_id]
        return the_handle

    @logExceptions()
    def release(self):
        with HandleWrapper.handle_lock:
            del HandleWrapper.handle_list[id(self)]


class VOFS(Operations):
    cacheTimeout = 30.0
    """
    The VOFS filesystem operations class.  Requires the vos (VOSpace)
    python package.

    To use this you will also need a VOSpace account from the CADC.
    """
    # disable operations VOSpace doesn't support
# MJG chown = None
# MJG link = None
    mknode = None
# MJG symlink = None
    getxattr = None
    listxattr = None
    removexattr = None

    def setxattr(self, path, name, value, options, position=0):
        logger.warning("Extended attributes not supported: {0} {1} {2} {3} {4}".format(path, name, value, options, position))

    def __init__(self, root, cache_dir, options, conn=None,
                 cache_limit=1024, cache_nodes=False,
                 cache_max_flush_threads=10, secure_get=False):
        """Initialize the VOFS.

        cache_limit is in MB.
        The style here is to use dictionaries to contain information
        about the Node.  The full VOSpace path is used as the Key for
        most of these dictionaries."""

        self.cache_nodes = cache_nodes

        # Standard attributes of the Node
        # Where in the file system this Node is currently located
        self.loading_dir = {}

        # Command line options.
        self.opt = options

        # What is the 'root' of the VOSpace? (eg vos:MyVOSpace)
        self.root = root

        # VOSpace is a bit slow so we do some caching.
        self.cache = Cache(cache_dir, cache_limit, False, VOFS.cacheTimeout,
                           maxFlushThreads=cache_max_flush_threads)

        # All communication with the VOSpace goes through this client
        # connection.
        try:
            self.client = vos.Client(root_node=root, conn=conn) # , MJG
#                                     transfer_shortcut=True, secure_get=secure_get)
        except Exception as e:
            e = FuseOSError(getattr(e, 'errno', EIO))
            e.filename = root
            e.strerror = getattr(e, 'strerror', 'failed while making mount')
            raise e

        # Create a condition variable to get rid of those nasty sleeps
        self.condition = CacheCondition(lock=None, timeout=VOFS.cacheTimeout)

    def __call__(self, op, *args):
        logger.debug('-> {0} {1}'.format(op, repr(args)))
        ret = None
        try:
            if not hasattr(self, op):
                raise FuseOSError(EFAULT)
            ret = getattr(self, op)(*args)
            return ret
        except Exception as all_exceptions:
            ret = str(all_exceptions)
            errno = getattr(all_exceptions, 'errno', None)
            errno = errno is not None and errno or EAGAIN
            exception = FuseOSError(errno)
            raise exception
        finally:
            logger.debug('<- {0} {1}'.format(op, repr(ret)))

    @logExceptions()
    def access(self, path, mode):
        """Check if path is accessible.

        Only checks read access, mode is currently ignored"""
        logger.debug("Checking if -->{0}<-- is accessible".format(path))
        try:
            self.getNode(path)
        except:
            return -1
        return 0

    @logExceptions()
    def chmod(self, path, mode):
        """
        Set the read/write groups on the VOSpace node based on chmod style
        modes.

        This function is a bit funny as the VOSpace spec sets the name
        of the read and write groups instead of having mode setting as
        a separate action.  A chmod that adds group permission thus
        becomes a chgrp action.

        Here I use the logic that the new group will be inherited from
        the container group information.
        """
        logger.debug("Changing mode for %s to %d" % (path, mode))

        node = self.getNode(path)
        parent = self.getNode(os.path.dirname(path))

# MJG    if node.groupread == "NONE":
        if node.groupread is None:
            node.groupread = parent.groupread
#        if node.groupwrite == "NONE":
        if node.groupwrite is None:
            node.groupwrite = parent.groupwrite
        # The 'node' object returned by getNode has a chmod method
        # that we now call to set the mod, since we set the group(s)
        # above.  NOTE: If the parrent doesn't have group then NONE is
        # passed up and the groupwrite and groupread will be set tovos:
        # the string NONE.
        if node.chmod(mode):
            # Now set the time of change/modification on the path...
            # TODO: This has to be broken. Attributes may come from Cache if
            # the file is modified. Even if they don't come from the cache,
            # the getAttr method calls getNode with force=True, which returns
            # a different Node object than "node". The st_ctime value will be
            # updated on the newly replaced node in self.node[path] but
            # not in node, then node is pushed to vospace without the st_time
            # change, and then it is pulled back, overwriting the change that
            # was made in self.node[path]. Plus modifying the mtime of the file
            # is not conventional Unix behaviour. The mtime of the parent
            # directory would be changed.
            self.getattr(path)['st_ctime'] = time.time()
            # if node.chmod returns false then no changes were made.
            self.client.update(node)
            self.getNode(path, force=True)

    @logExceptions()
    def chown(self, path, uid, gid):
        """Change the permissions. Currently does nothing since this
        functionality is not present in VOSpace. Dummy implementation
        here to prevent error messages from file compression codes.
        """
        pass

    @logExceptions()
    def create(self, path, flags, fi=None):
        """Create a node. Currently ignores the ownership mode
        @param path: the container/dataNode in VOSpace to be created
        @param flags: Read/Write settings (eg. 600)
        """

        logger.debug("Creating a node: {0} with flags {1}".format(path, str(flags)))

        # Create is handle by the client.
        # This should fail if the base path doesn't exist
        self.client.open(path, os.O_CREAT).close()

        node = self.getNode(path)
        parent = self.getNode(os.path.dirname(path))
        # Force inheritance of group settings.
        node.groupread = parent.groupread
        node.groupwrite = parent.groupwrite
        if node.chmod(flags):
            # chmod returns True if the mode changed but doesn't do update.
            self.client.update(node)
            self.getNode(path, force=True)

        # now we can just open the file in the usual way and return the handle
        return self.open(path, os.O_WRONLY)

    def destroy(self, path):
        """Called on filesystem destruction. Path is always /

           Call the flushNodeQueue join() method which will block
           until any running and/or queued jobs are finished"""

        if self.cache.flushNodeQueue is None:
            raise CacheError("flushNodeQueue has not been initialized")
        self.cache.flushNodeQueue.join()
        self.cache.flushNodeQueue = None

    @logExceptions()
    def fsync(self, path, data_sync, file_id):
        if self.opt.readonly:
            logger.debug("File system is readonly, no sync allowed")
            return

        try:
            fh = HandleWrapper.file_handle(file_id)
        except KeyError:
            raise FuseOSError(EIO)

        if fh.read_only:
            raise FuseOSError(EPERM)

        fh.cache_file_handle.fsync()

    def get_node(self, path, force=False, limit=0):
        """Use the client and pull the node from VOSpace.

        Currently force=False is the default... so we never check
        VOSpace to see if the node metadata is different from what we
        have.  This doesn't keep the node metadata current but is
        faster if VOSpace is slow.
        @type limit: int or None
        @rtype : vos.Node
        @param path: the VOSpace node to get
        @param force: force retrieval (true) or provide cached version if available (false)?
        @param limit: Number of child nodes to retrieve per request, if limit  is None then get max returned by service.
        """

        # Pull the node meta data from VOSpace.
        logger.debug("requesting node {0} from VOSpace. Force: {1}".format(path, force))
        node = self.client.get_node(path, force=force, limit=limit)
        logger.debug("Got node {0}".format(node.name))
        return node

    getNode = get_node

    # @logExceptions()
    def getattr(self, path, file_id=None):
        """
        Build some attributes for this file, we have to make-up some stuff
        """
        # Try to get the attributes from the cache first. This will only return
        # a result if the files has been modified and not flushed to vospace.
        attr = self.cache.getAttr(path)
        return attr is not None and attr or self.get_node(path, limit=0, force=False).attr

    def init(self, path):
        """Called on filesystem initialization. (Path is always /)

        Here is where we start the worker threads for the queue that flushes nodes.
        """
        self.cache.flushNodeQueue = FlushNodeQueue(maxFlushThreads=self.cache.maxFlushThreads)

    # MJG
    # @logExceptions()
    def link(self, target, name):
        """Create a link node in the VOSpace at the specified location.
        """
        try:
            self.client.link(target, name)
        except OSError as os_error:
            raise FuseOSError(getattr(os_error, 'errno', EFAULT))

    # @logExceptions()
    def symlink(self, name, target):
        """Create a link node in the VOSpace at the specified location.
        There is no distinction in VOSpace between hard and symbolic links.
        """
        try:
            if target[0] != '/': target = '/' + target # MJG URL hack 
            self.client.link(target, name)
        except OSError as os_error:
            raise FuseOSError(getattr(os_error, 'errno', EFAULT))

    # @logExceptions()
    def mkdir(self, path, mode):
        """Create a container node in the VOSpace at the correct location.

        set the mode after creation. """
        try:
            self.client.mkdir(path)
        except OSError as os_error:
            if "read-only mode" in str(os_error):
                raise FuseOSError(EPERM)
            raise FuseOSError(getattr(os_error, 'errno', EFAULT))
        # self.chmod(path, mode)

    # @logExceptions()
    def open(self, path, flags, *mode):
        """Open file with the desired modes

        Here we return a handle to the cached version of the file
        which is updated if older and out of sync with VOSpace.

        """

        logger.debug("Opening %s with flags %s" % (path, flag2mode(flags)))
        node = None

        # according to man for open(2), flags must contain one of O_RDWR,
        # O_WRONLY or O_RDONLY. Because O_RDONLY=0 and options other than
        # O_RDWR, O_WRONLY and O_RDONLY may be present,
        # readonly = (flags == O_RDONLY) and readonly = (flags | O_RDONLY)
        # won't work. The only way to detect if it's a read only is to check
        # whether the other two flags are absent.
        read_only = ((flags & (os.O_RDWR | os.O_WRONLY)) == 0)

        must_exist = not ((flags & os.O_CREAT) == os.O_CREAT)
        cache_file_attrs = self.cache.getAttr(path)
        if cache_file_attrs is None and not read_only:
            # file in the cache not in the process of being modified.
            # see if this node already exists in the cache; if not get info
            # from vospace
            try:
                node = self.get_node(path)
            except OSError as e:
                if e.errno == 404:
                    # file does not exist
                    if not flags & os.O_CREAT:
                        # file doesn't exist unless created
                        raise FuseOSError(ENOENT)
                else:
                    raise FuseOSError(e.errno)

        # check if this file is locked, if locked on vospace then don't open
        locked = False

        if node and node.props.get('islocked', False):
            logger.debug("%s is locked." % path)
            locked = True

        if not read_only and node and not locked:
            if node.type == "vos:DataNode":
                parent_node = self.get_node(os.path.dirname(path), force=False, limit=1)
                if parent_node.props.get('islocked', False):
                    logger.debug("%s is locked by parent node." % path)
                    locked = True
            elif node.type == "vos:LinkNode":
                try:
                    # sometimes target_nodes aren't internal... so then not
                    # locked
                    target_node = self.get_node(node.target, force=False, limit=1)
                    if target_node.props.get('islocked', False):
                        logger.debug("{0} target node is locked.".format(path))
                        locked = True
                    else:
                        target_parent_node = self.get_node(os.path.dirname(node.target), force=False, limit=1)
                        if target_parent_node.props.get('islocked', False):
                            logger.debug("{0} parent node is locked.".format(path))
                            locked = True
                except Exception as lock_exception:
                    logger.warn("Error while checking for lock: {0}".format(str(lock_exception)))
                    pass

        if locked and not read_only:
            # file is locked, cannot write
            e = OSError(EPERM)
            e.filename = path
            e.strerror = "Cannot write to locked file"
            logger.debug("{0}".format(e))
            raise e

        my_proxy = MyIOProxy(self, path)
        if node is not None:
            my_proxy.set_size(int(node.props.get('length')))
            my_proxy.set_md5(node.props.get('MD5'))

        logger.debug("IO Proxy initialized:{0}  in backing.".format(my_proxy))

        # new file in cache library or if no node information (node not in vospace).
        handle = self.cache.open(path, flags & os.O_WRONLY != 0, must_exist, my_proxy, self.cache_nodes)

        logger.debug("Creating file:{0}  in backing.".format(path))

        if flags & os.O_TRUNC != 0:
            handle.truncate(0)
        if node is not None:
            handle.setHeader(my_proxy.getSize(), my_proxy.get_md5())
        return HandleWrapper(handle, read_only).get_id()

    @logExceptions()
    def read(self, path, size=0, offset=0, file_id=None):
        """
        Read the required bytes from the file and return a buffer containing
        the bytes.
        """

        # Read from the requested file_handle, which was set during 'open'
        if file_id is None:
            raise FuseOSError(EIO)

        logger.debug("reading range: %s %d %d %d" % (path, size, offset, file_id))

        while True:
            try:
                fh = HandleWrapper.file_handle(file_id)
            except KeyError as e:
                logger.debug(str(e))
                raise FuseOSError(EIO)
            with self.condition:
                bytes = fh.cache_file_handle.read(size, offset)
            # Send back if we got bytes or we've read to end of file already.
            if len(bytes) > 0 or not fh.cache_file_handle.metaData.size - offset > 0:
                break
        return bytes

    @logExceptions()
    def readlink(self, path):
        """
        Return a string representing the path to which the symbolic link
        points.

        path: filesystem location that is a link

        returns the file that path is a link to.

        Currently doesn't provide correct capabilty for VOSpace FS.
        """
        return self.get_node(path).name+"?link="+urllib.quote_plus(self.getNode(path).target)

    @logExceptions()
    def readdir(self, path, file_id):
        """Send a list of entries in this directory"""
        logger.debug("Getting directory list for {0}".format(path))
        # reading from VOSpace can be slow, we'll do this in a thread
        import thread
        with self.condition:
            if not self.loading_dir.get(path, False):
                self.loading_dir[path] = True
                thread.start_new_thread(self.load_dir, (path, ))

            while self.loading_dir.get(path, False):
                logger.debug("Waiting ... ")
                self.condition.wait()
        return ['.', '..'] + [e.name.encode('utf-8') for e in self.getNode(path,
                                                                           force=False,
                                                                           limit=None).node_list]

    @logExceptions()
    def load_dir(self, path):
        """Load the dirlist from VOSpace.

        This should always be run in a thread."""
        try:
            logger.debug("Starting getNodeList thread")
            node_list = self.getNode(path, force=True, limit=None).node_list
            logger.debug("Got listing {0} for {1}".format(node_list, path))
        finally:
            self.loading_dir[path] = False
            with self.condition:
                self.condition.notify_all()
        return

    @logExceptions()
    def flush(self, path, file_id):

        logger.debug("flushing {0}".format(file_id))
        fh = HandleWrapper.file_handle(file_id)

        try:
            fh.cache_file_handle.flush()
        except CacheRetry as ce:
            logger.critical(str(ce))
            logger.critical("Push to VOSpace reached FUSE timeout, continuing VOSpace push in background.")
            pass
        return 0

    @logExceptions()
    def release(self, path, file_id):
        """Close the file.

        @param path: vospace path to close reference to
        @param file_id: file_handle_id to close reference to
        """
        logger.debug("releasing file %d " % file_id)
        fh = HandleWrapper.file_handle(file_id)
        fh.cache_file_handle.release()
        if fh.cache_file_handle.fileModified:
            # This makes the node disappear from the nodeCache.
            with self.client.nodeCache.volatile(path):
                pass
        return fh.release()

    @logExceptions()
    def rename(self, src, dest):
        """Rename a data node into a new container"""
        logger.debug("Original %s -> %s" % (src, dest))
        try:
            logger.debug("Moving %s to %s" % (src, dest))
            result = self.client.move(src, dest)
            logger.debug(str(result))
            if result:
                self.cache.renameFile(src, dest)
                return 0
            return -1
        except Exception as e:
            logger.error("%s" % str(e))
            import re
            if re.search('NodeLocked', str(e)) is not None:
                raise OSError(EPERM)
            raise

    @logExceptions()
    def rmdir(self, path):
        node = self.getNode(path)
        if node and node.props.get('islocked', False):
            logger.debug("%s is locked." % path)
            raise FuseOSError(EPERM)
        self.client.delete(path)

    @logExceptions()
    def statfs(self, path):
        node = self.getNode(path)
        block_size = 512
        n_bytes = 2 ** 33
        free = 2 ** 33

        if 'quota' in node.props:
            n_bytes = int(node.props.get('quota', 2 ** 33))
            used = int(node.props.get('length', 2 ** 33))
            free = n_bytes - used
        logger.debug("Got properties: {0}".format(node.props))
        sfs = {'f_bsize': block_size, 'f_frsize': block_size, 'f_blocks': int(n_bytes / block_size),
               'f_bfree': int(free / block_size), 'f_bavail': int(free / block_size),
               'f_files': len(node.node_list), 'f_ffree': 2 * 10, 'f_favail': 2 * 10, 'f_flags': 0,
               'f_namemax': 256}
        return sfs

    @logExceptions()
    def truncate(self, path, length, file_id=None):
        """Perform a file truncation to length bytes"""
        logger.debug("Attempting to truncate {0}({1}) to length {2}".format(path, file_id, length))

        if file_id is None:
            # Ensure the file exists.
            if length == 0:
                fh = self.open(path, os.O_RDWR | os.O_TRUNC)
                self.release(path, fh)
                return

            fh = self.open(path, os.O_RDWR)
            try:
                self.truncate(path, length, fh)
            finally:
                self.release(path, fh)
        else:
            fh = HandleWrapper.file_handle(file_id)
            if fh.read_only:
                logger.debug("file was not opened for writing")
                raise FuseOSError(EPERM)
            fh.cache_file_handle.truncate(length)
        return

    @logExceptions()
    def unlink(self, path):
        node = self.getNode(path, force=False, limit=1)
        if node and node.props.get('islocked', False):
            logger.debug("%s is locked." % path)
            raise FuseOSError(EPERM)
        self.cache.unlinkFile(path)
        if node:
            self.client.delete(path)

    @logExceptions()
    def write(self, path, data, size, offset, file_id=None):

        if self.opt.readonly:
            logger.debug("File system is readonly.. so writing 0 bytes\n")
            return 0

        try:
            fh = HandleWrapper.file_handle(file_id)
        except KeyError:
            raise FuseOSError(EIO)

        if fh.read_only:
            logger.debug("file was not opened for writing")
            raise FuseOSError(EPERM)

        logger.debug("%s -> %s" % (path, fh))
        logger.debug("%d --> %d" % (offset, offset + size))
        return fh.cache_file_handle.write(data, size, offset)



#!/usr/bin/env python
#
# AUTHCLIENT -- Client methods for the Data Lab Authentication Service
#

from __future__ import print_function

__authors__ = 'Mike Fitzpatrick <fitz@noao.edu>, Data Lab <datalab@noao.edu>'
__version__ = '20190422'  # yyyymmdd


'''
    Client methods for the Data Lab Authentication Manager Service.

    Auth Manager Client Interface
    -----------------------------

                 login  (user, password=None, debug=False, verbose=False)
                logout  (token=None)

                whoAmI  ()
               isAlive  (svc_url=DEF_SERVICE_URL)
          isValidToken  (token)
           isValidUser  (user)
       isValidPassword  (user, password)
             hasAccess  (user, resource)
        isUserLoggedIn  (user)
       isTokenLoggedIn  (token)
         passwordReset  (token, username, password)

           set_svc_url  (svc_url)
           get_svc_url  ()
           set_profile  (profile)
           get_profile  ()
             getClient  ()


Import via

.. code-block:: python

    from dl import authClient
'''

import requests
import socket
import os
import sys
from time import gmtime, strftime

try:
    from urllib import urlencode                # Python 2
    import ConfigParser                         # Python 2
except ImportError:
    from urllib.parse import urlencode          # Python 3
    import configparser as ConfigParser         # Python 3

#if os.path.isfile('./Util.py'):                # use local dev copy
#    from Util import def_token
#else:                                           # use distribution copy
#    from dl.Util import def_token
from dl.Util import def_token

is_py3 = sys.version_info.major == 3



# Pre-defined authentication tokens. These are fixed strings that provide
# limited access to Data Lab services, this access is controlled on the
# server-side so we don't need strict security here.

ANON_TOKEN = "anonymous.0.0.anon_access"
DEMO_TOKEN = "dldemo.99999.99999.demo_access"
TEST_TOKEN = "dltest.99998.99998.test_access"

# Set the default user accounts for the authentication service.  We don't
# include privileged users so that account can remain secure.

DEF_USERS = {'anonymous': ANON_TOKEN,
             'dldemo': DEMO_TOKEN,
             'dltest': TEST_TOKEN}


# The URL of the AuthManager service to contact.  This may be changed by
# passing a new URL into the set_svc_url() method before beginning.

DEF_SERVICE_ROOT = "https://datalab.noao.edu"

# Allow the service URL for dev/test systems to override the default.
THIS_HOST = socket.gethostname()
if THIS_HOST[:5] == 'dldev':
    DEF_SERVICE_ROOT = "http://dldev.datalab.noao.edu"
elif THIS_HOST[:6] == 'dltest':
    DEF_SERVICE_ROOT = "http://dltest.datalab.noao.edu"

DEF_SERVICE_URL = DEF_SERVICE_ROOT + "/auth"
SM_SERVICE_URL  = DEF_SERVICE_ROOT + "/storage"
QM_SERVICE_URL  = DEF_SERVICE_ROOT + "/query"


# The requested authentication "profile".  A profile refers to the specific
# machines and services used by the AuthManager on the server. [Note that
# profiles are not currently used.]

DEF_SERVICE_PROFILE = "default"

# Use a /tmp/AM_DEBUG file as a way to turn on debugging in the client code.
DEBUG = os.path.isfile('/tmp/AM_DEBUG')


# ######################################################################
#
#  Authentication Client Interface
#
#  This API provides convenience methods that allow an application to
#  import the Client class without having to explicitly instantiate a
#  class object.  The parameter descriptions and example usage is given
#  in the comments for the class methods.  Module methods have their
#  docstrings patched below.
#
# ######################################################################


# ###################################
#  Authentication error class
# ###################################

class dlAuthError(Exception):
    '''A throwable error class.
    '''
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


# User methods -- All methods except login() return either a 'True' string
# or an error of the form 'ERR <message>'.  On success, the login() method
# will return the user-id token.

def login(user, password=None, debug=False, verbose=False):
    response = 'OK'
    if user in list(DEF_USERS.keys()):
        return DEF_USERS[user]
    else:
        try:
            response = ac_client.login(user, password, debug=debug,
                                       verbose=verbose)
        except Exception as e:
            response = str(e)
    return response


def whoAmI():
    user = 'anonymous'
    try:
        token = def_token(None)
        user, uid, gid, hash = token.strip().split('.', 3)
    except:
        return 'anonymous'
    else:
        return user

def isAlive(svc_url=DEF_SERVICE_URL):
    try:
        response = ac_client.isAlive(svc_url.strip('/'))
    except Exception as e:
        response = str(e)
    return (True if response.lower() == 'true' else False)


def isValidToken(token):
    try:
        user, uid, gid, hash = token.strip().split('.', 3)
    except Exception as e:
        return False

    if user in list(DEF_USERS.keys()) and token in list(DEF_USERS.values()):
        return True
    else:
        try:
            response = ac_client.isValidToken(token)
        except Exception:
            return False
    return (True if response.lower() == 'true' else False)


def isValidUser(user):
    if user in list(DEF_USERS.keys()):
        return True
    else:
        try:
            response = ac_client.isValidUser(user)
        except Exception as e:
            response = str(e)
    return (True if response.lower() == 'true' else False)


def isValidPassword(user, password):
    if (user == password) and (user in list(DEF_USERS.keys())):
        return True
    else:
        try:
            response = ac_client.isValidPassword(user, password)
        except Exception as e:
            response = str(e)
    return (True if response.lower() == 'true' else False)


def hasAccess(user, resource):
    try:
        response = ac_client.hasAccess(user, resource)
    except Exception as e:
        response = str(e)
    return (True if response.lower() == 'true' else False)


def isUserLoggedIn(user):
    try:
        response = ac_client.isUserLoggedIn(user)
    except Exception as e:
        response = str(e)
    return (True if response.lower() == 'true' else False)


def isTokenLoggedIn(token):
    try:
        response = ac_client.isTokenLoggedIn(token)
    except Exception as e:
        response = str(e)
    return (True if response.lower() == 'true' else False)


def logout(token=None):
    try:
        response = ac_client.logout(def_token(token))
    except Exception as e:
        response = str(e)
    return response


def passwordReset(token, username, password):
    try:
        response = ac_client.passwordReset(token, username, password)
    except Exception as e:
        response = str(e)
    return response


# Standard Service Methods
def set_svc_url(svc_url):
    return ac_client.set_svc_url(svc_url.strip('/'))


def get_svc_url():
    return ac_client.get_svc_url()


def set_profile(profile):
    return ac_client.set_profile(profile)


def get_profile():
    return ac_client.get_profile()


def list_profiles(token, profile=None, format='text'):
    return "None"



#####################################
#  Authentication client procedures
#####################################

class authClient(object):
    '''
         AUTHCLIENT -- Client-side methods to access the Data Lab
                       Authentication Service.
    '''

    def __init__(self):
        '''Initialize the authClient class. '''
        self.svc_url = DEF_SERVICE_URL          # service URL
        self.svc_profile = DEF_SERVICE_PROFILE  # service prfile
        self.username = ""                      # default client logn user
        self.auth_token = None                  # default client logn token

        # Check the $HOME/.datalab directory for a valid token.  If that dir
        # doesn't already exist, create it so we can store the new token.
        self.home = '%s/.datalab' % os.path.expanduser('~')
        if not os.path.exists(self.home):
            os.makedirs(self.home)
        self.loadConfig()                       # load config file

        self.debug = DEBUG                      # interface debug flag

    def loadConfig(self):
        '''Read the $HOME/.datalab/dl.conf file.
        '''
        self.config = ConfigParser.RawConfigParser(allow_no_value=True)
        if os.path.exists('%s/dl.conf' % self.home):
            self.config.read('%s/dl.conf' % self.home)
        else:
            self.config.add_section('datalab')
            self.config.set('datalab', 'created', strftime(
                '%Y-%m-%d %H:%M:%S', gmtime()))
            self.config.add_section('login')
            self.config.set('login', 'status', 'loggedout')
            self.config.set('login', 'user', '')
            self.config.set('login', 'authtoken', '')

            self.config.add_section('auth')
            self.config.set('auth', 'profile', 'default')
            self.config.set('auth', 'svc_url', DEF_SERVICE_URL)

            self.config.add_section('query')
            self.config.set('query', 'profile', 'default')
            self.config.set('query', 'svc_url', QM_SERVICE_URL)

            self.config.add_section('storage')
            self.config.set('storage', 'profile', 'default')
            self.config.set('storage', 'svc_url', SM_SERVICE_URL)

            self.config.add_section('vospace')
            self.config.set('vospace', 'mount', '')

            self.writeConfig()

    def setConfig(self, section, param, value):
        '''Set a value and save the configuration file.
        '''
        if not self.config.has_section(section):
            self.config.add_section(section)
        self.config.set(section, param, value)
        self.writeConfig()

    def getConfig(self, section, param):
        '''Get a value from the configuration file.
        '''
        return self.config.get(section, param)

    def writeConfig(self):
        '''Write out the configuration file to disk.
        '''
        with open('%s/dl.conf' % self.home, 'w') as configfile:
            self.config.write(configfile)

    def whoAmI():
        '''Return the currently logged-in user identity.

        Parameters
        ----------
        None

        Returns
        -------
        name : str
            Currently logged-in user name, or 'anonymous' if not logged-in

        Example
        -------
        .. code-block:: python

            from dl import authClient
            name = authClient.whoAmI()
        '''
        user = 'anonymous'
        try:
            token = def_token(None)
            user, uid, gid, hash = token.strip().split('.', 3)
        except:
            return 'anonymous'
        else:
            return user


    # Standard Data Lab service methods.
    #
    def set_svc_url(self, svc_url):
        '''Set the URL of the Authentication Service to be used.

        Parameters
        ----------
        svc_url : str
            Authentication service base URL to call.

        Returns
        -------
        Nothing

        Example
        -------
        .. code-block:: python

            from dl import authClient
            authClient.set_svc_url("http://localhost:7001/")
        '''
        self.svc_url = acToString(svc_url.strip('/'))

    def get_svc_url(self):
        '''Return the currently-used Authentication Service URL.

        Parameters
        ----------
        None

        Returns
        -------
        service_url : str
            The currently-used Authentication Service URL.

        Example
        -------
        .. code-block:: python

            from dl import authClient
            service_url = authClient.get_svc_url()
        '''
        return acToString(self.svc_url)

    def set_profile(self, profile):
        '''Set the requested service profile.

        Parameters
        ----------
        profile : str
            Requested service profile string.

        Returns
        -------
        Nothing

        Example
        -------
        .. code-block:: python

            from dl import authClient
            token = authClient.client.set_profile("dev")
        '''
        self.svc_profile = acToString(profile)

    def get_profile(self):
        '''Get the requested service profile.

        Parameters
        ----------
        None

        Returns
        -------
        profile : str
            The currently requested service profile.

        Example
        -------
        .. code-block:: python

            from dl import authClient
            profile = authClient.client.get_profile()
        '''
        return acToString(self.svc_profile)

    def list_profiles(self, token, profile=None, format='text'):
        '''List the service profiles which can be accessed by the user.

        Parameters
        ----------
        token : str
            Valid auth service token.

        Returns
        -------
        profiles : JSON string

        Example
        -------
        .. code-block:: python

            from dl import authClient
            profiles = authClient.client.list_profiles(token, profile, format)
        '''
        pass

    def isAlive(self, svc_url=DEF_SERVICE_URL):
        '''Check whether the AuthManager service at the given URL is
            alive and responding.  This is a simple call to the root
            service URL or ping() method.

        Parameters
        ----------
        service_url : str
            The Query Service URL to ping.

        Returns
        -------
        result : bool
            True if service responds properly, False otherwise

        Example
        -------
        .. code-block:: python

            from dl import authClient
            if authClient.isAlive():
                print("Auth Manager is alive")
        '''
        url = svc_url
        try:
            r = requests.get(url, timeout=2)
            resp = r.text

            if r.status_code != 200:
                return False
            elif resp is not None and r.text.lower()[:11] != "hello world":
                return False
        except Exception:
            return False

        return True


    ###################################################
    #  SESSION MANAGEMENT
    ###################################################

    def login(self, username, password, debug=False, verbose=False):
        '''Authenticate the user with the Authentication Service.
            We first check for a valid login token in the user's
            $HOME/.datalab/ directory and simply return that rather
            than make a service call to get a new token.  If a token
            exists but is invalid, we remove it and get a new
            token.  In either case, we save the token for later use.

        Parameters
        ----------
        username : str
            User login name.

        password : str
            User password.  If not given, a valid ID token will be
            searched for in the $HOME/.datalab directory.

        debug : bool
            Method debug flag.
        verbose : bool
            Initialize session to print verbose output messages.

        Returns
        -------
        token : str
            One-time security token for valid user(identified via
            'username' and 'password').

        Example
        -------
        .. code-block:: python

            from dl import authClient
            token = authClient.login('dldemo', 'dldemo')   # get security token
        '''

        # Check the $HOME/.datalab directory for a valid token.  If that dir
        # doesn't already exist, create it so we can store the new token.
        if not os.path.exists(self.home):
            os.makedirs(self.home)

        # See if a datalab token file exists for the requested user.
        tok_file = ('%s/id_token.%s' % (self.home, username))
        if debug or self.debug:
            print("top of login: tok_file = '" + tok_file + "'")
            print("top of login: self.auth_token = '%s'" %
                   str(self.auth_token))
            print("top of login: token = ")
            os.system("cat " + tok_file)

        if password is None:
            if os.path.exists(tok_file) and os.stat(tok_file).st_size > 0:
                with open(tok_file, "r") as tok_fd:
                    o_tok = acToString(tok_fd.read(128))  # read the old token

                # Return a valid token, otherwise remove the file and obtain a
                # new one.
                if o_tok.startswith(username+'.') and self.isValidToken(o_tok):
                    self.username = username
                    self.auth_token = o_tok
                    if debug or self.debug:
                        print("using old token for '%s'" % username)
                    return acToString(o_tok)
                else:
                    if debug or self.debug:
                        print("removing invalid token file '%s'" % tok_file)
                    os.remove(tok_file)

        # Either the user is not logged in or the token is invalid, so
        # make a service call to get a new token.
        url = self.svc_url + "/login?"
        query_args = {"username": username,
                      "password": password,
                      "profile": self.svc_profile,
                      "debug": (debug or self.debug)}

        response = 'None'
        try:
            r = requests.get(url, params=query_args)
            response = acToString(r.content)

            if debug or self.debug:
                print("%s:  resp = '%s'" % (str(r.status_code),response))
            if r.status_code != 200:
                raise Exception(response)

        except Exception as e:
            if debug or self.debug:
                print("Raw exception msg = '%s'" % acToString(r.content))
            if self.isAlive(self.svc_url) == False:
                raise dlAuthError("AuthManager Service not responding.")
            if self.isValidUser(username):
                if password is None:
                    if not os.path.exists(tok_file):
                        raise dlAuthError("No password or token supplied")
                    else:
                        raise dlAuthError("No password supplied")
                elif not self.isValidPassword(username, password):
                    raise dlAuthError("Invalid password in login()")
                else:
                    raise dlAuthError(str(e))
            else:
                raise dlAuthError("Invalid username in login()")

        else:
            self.auth_token = response
            self.username = username

        # Save the token and config file.
        if os.access(self.home, os.W_OK):
            tok_file = '%s/id_token.%s' % (self.home, username)
            with open(tok_file, 'w') as tok_fd:
                if debug or self.debug:
                    print("login: writing new token for '%s'" % username)
                    print("login: self.auth_token = '%s'" %
                           str(self.auth_token))
                    print("login: token = ")
                    os.system('cat ' + tok_file)

                tok_fd.write(self.auth_token)

            self.config.set('login', 'status', 'loggedin')
            self.config.set('login', 'user', username)
            self.config.set('login', 'authtoken', self.auth_token)
            self.writeConfig()

        return acToString(self.auth_token)

    def logout(self, token=None):
        '''Log the user out of the Data Lab.

        Parameters
        ----------
        token : str
            User login token

        Returns
        -------
        'OK' string on success, or exception message

        Example
        -------
        .. code-block:: python

            from dl import authClient
            status = authClient.logout(token)
        '''
        url = self.svc_url + "/logout?"
        args = urlencode({"token": token,
                          "debug": self.debug})
        url = url + args

        if self.debug:
            print("logout: token = '%s'" % token)
            print("logout: auth_token = '%s'" % self.auth_token)
            print("logout: url = '%s'" % url)
            print("logout: logged in = " + self.isTokenLoggedIn(token))

        try:
            user, uid, gid, hash = token.strip().split('.', 3)
        except Exception as e:
            raise dlAuthError('Error: Invalid user token')

        if not isValidToken(token):
            raise dlAuthError("Error: Invalid user token")
        if not isTokenLoggedIn(token):
            raise dlAuthError("Error: User token is not logged in")

        try:
            # Add the auth token to the reauest header.
            headers = {'X-DL-AuthToken': token}

            r = requests.get(url, params=args, headers=headers)
            response = acToString(r.content)

            if r.status_code != 200:
                raise Exception(response)

        except Exception as e:
            raise dlAuthError(str(e))
        else:
            self.auth_token = None
            tok_file = self.home + '/id_token.' + user
            if os.path.exists(tok_file):
                os.remove(tok_file)

            self.config.set('login', 'status', 'loggedout')
            self.config.set('login', 'user', '')
            self.config.set('login', 'authtoken', '')
            self.writeConfig()

        return response

    def passwordReset(self, token, username, password):
        '''Reset a user password reset.  We require that the user provide
           either a valid 'root' token or the token for the account being
           reset.

        Parameters
        ----------
        token : str
            User login token
        username : str
            User login name
        password : str
            User password

        Returns
        -------
        'OK' string on success, or exception message

        Example
        -------
        .. code-block:: python

            from dl import authClient
            status = authClient.logout(token)
        '''
        url = self.svc_url + "/passwordReset?"
        args = urlencode({"token": token,
                          "username": username,
                          "debug": self.debug})

        if self.debug:
            print("passwdReset: token = '%s'" % token)
            print("passwdReset: auth_token = '%s'" % self.auth_token)
            print("passwdReset: url = '%s'" % url)

        if not self.isValidToken(token):
            if self.debug:
                print("passwdReset: Invalid user token")
            raise Exception("Error: Invalid user token")

        # Reset the auth_token to the one passed in by the service call.
        self.auth_token = token

        user, uid, gid, hash = self.auth_token.strip().split('.', 3)
        if user != 'root' and user != username:
            raise Exception("Error: Invalid user or non-root token")

        try:
            # Add the auth token and password to the reauest header.
            headers = {'X-DL-AuthToken': token,
                       'X-DL-Password': password}

            r = requests.get(url, params=args, headers=headers)
            response = acToString(r.content)

            if r.status_code != 200:
                raise Exception(r.content)

        except Exception:
            raise dlAuthError(acToString(r.content))
        else:
            # Update the saved user token.
            if response is not None:
                self.auth_token = response
                tok_file = self.home + '/id_token.' + self.username
                if os.path.exists(tok_file):
                    os.remove(tok_file)
                with open(tok_file, 'wb') as tok_fd:
                    if self.debug:
                        print("pwreset: writing new token for '%s'" + username)
                        print("pwreset: response = '%s'" + response)
                        print("pwreset: token = '%s'" + self.auth_token)
                    tok_fd.write(acToString(self.auth_token))
                    tok_fd.close()
            else:
                print('pwReset response is None')

        return response

    def hasAccess(self, token, resource):
        '''See whether the given token has access to the named Resource.

        Parameters
        ----------
        token : str
            User login token
        resource : str
            Resource identifier to check

        Returns
        -------
        status : bool
            True if user owns or has access, False otherwise

        Example
        -------
        .. code-block:: python

            from dl import authClient
            status = authClient.hasAccess(token,'vos://test.dat')
        '''
        # Either the user is not logged in or the token is invalid, so
        # make a service call to get a new token.
        url = self.svc_url + "/hasAccess?"
        args = urlencode({"user": token,
                          "resource": resource,
                          "profile": self.svc_profile})
        url = url + args

        if self.debug:
            print("hasAccess: url = '%s'" % url)

        return self.retBoolValue(url)

    def isValidToken(self, token):
        '''See whether the current token is valid.

        Parameters
        ----------
        token : str
            User login token

        Returns
        -------
        status : bool
            True if token is valid, False otherwise

        Example
        -------
        .. code-block:: python
            if authClient.isValidToken(token):
                print("Valid token")
        '''
        url = self.svc_url + "/isValidToken?"
        args = urlencode({"token": token,
                          "profile": self.svc_profile})
        url = url + args

        if self.debug:
            print("isValidToken: url = '%s'" % url)

        # Save the value before returning so we can print it in debug mode.
        isValid = self.retBoolValue(url)
        if self.debug:
            print("isValidToken: valid = " + str(isValid))

        return isValid

    def isValidPassword(self, user, password):
        '''See whether the password is valid for the user.

        Parameters
        ----------
        user : str
            User login name
        password : str
            Password for named user

        Returns
        -------
        status : bool
            True if password is valid for the user, False otherwise

        Example
        -------
        .. code-block:: python
            if authClient.isValidPassword('monty','python'):
                print("Valid password")
        '''
        url = self.svc_url + "/isValidPassword?"
        args = urlencode({"user": user,
                          "password": password,
                          "profile": self.svc_profile})
        url = url + args

        if self.debug:
            print("isValidPassword: url = '%s'" % url)

        try:
            val = self.retBoolValue(url)
        except Exception:
            val = "False"

        return val

    def isValidUser(self, user):
        '''See whether the specified user is valid.

        Parameters
        ----------
        user : str
            User login name

        Returns
        -------
        status : bool
            True if 'user' is a valid user name, False otherwise

        Example
        -------
        .. code-block:: python
            if authClient.isValidUser('monty'):
                print("Valid user")
        '''
        url = self.svc_url + "/isValidUser?"
        args = urlencode({"user": user,
                          "profile": self.svc_profile})
        url = url + args

        if self.debug:
            print("isValidUser: url = '%s'" % url)

        try:
            val = self.retBoolValue(url)
        except Exception:
            val = "False"

        return val

    def isUserLoggedIn(self, user):
        '''See whether the user identified by the token is currently logged in.

        Parameters
        ----------
        user : str
            User login name

        Returns
        -------
        status : bool
            True if user is currently logged-in, False otherwise

        Example
        -------
        .. code-block:: python
            if not authClient.isUserLoggedIn(token):
                token = authClient.login('monty')
        '''
        url = self.svc_url + "/isUserLoggedIn?"
        args = urlencode({"user": user,
                          "profile": self.svc_profile})
        url = url + args

        if self.debug:
            print("isUserLoggedIn: url = '%s'" % url)

        try:
            val = self.retBoolValue(url)
        except Exception:
            val = "False"

        return val

    def isTokenLoggedIn(self, token):
        '''See whether the user identified by the token is currently logged in.

        Parameters
        ----------
        token : str
            User login token

        Returns
        -------
        status : bool
            True if token is marked as logged-in, False otherwise

        Example
        -------
        .. code-block:: python
            if not authClient.isTokenLoggedIn(token):
                token = authClient.login('monty')
        '''
        url = self.svc_url + "/isTokenLoggedIn?"
        args = urlencode({"token": token,
                          "profile": self.svc_profile})
        url = url + args

        if self.debug:
            print("isTokenLoggedIn: tok = '%s'" % token)
            print("isTokenLoggedIn: url = '%s'" % url)

        try:
            val = self.retBoolValue(url)
        except Exception:
            val = "False"

        return val


    ###################################################
    #  PRIVATE UTILITY METHODS
    ###################################################

    def debug(self, debug_val):
        '''Toggle debug flag.
        '''
        self.debug = debug_val

    def retBoolValue(self, url):
        '''Utility method to call a boolean service at the given URL.
        '''
        response = ""
        try:
            # Add the auth token to the reauest header.
            if self.auth_token != None:
                headers = {'X-DL-AuthToken': self.auth_token}
                r = requests.get(url, headers=headers)
            else:
                r = requests.get(url)
            response = acToString(r.content)

            if r.status_code != 200:
                raise Exception(r.content)

        except Exception:
            return acToString(r.content)
        else:
            return response

    def getHeaders(self, token):
        '''Get default tracking headers.
        '''
        tok = def_token(token)
        user, uid, gid, hash = tok.strip().split('.', 3)
        hdrs = {'Content-Type': 'text/ascii',
                'X-DL-ClientVersion': __version__,
                'X-DL-OriginIP': self.hostip,
                'X-DL-OriginHost': self.hostname,
                'X-DL-User': user,
                'X-DL-AuthToken': tok}                  # application/x-sql
        return hdrs

    def getFromURL(self, svc_url, path, token):
        '''Get something from a URL.  Return a 'response' object.
        '''
        try:
            hdrs = self.qc_getHeaders(token)
            resp = requests.get("%s%s" % (svc_url, path), headers=hdrs)

        except Exception as e:
            raise dlAuthError(str(e))
        return resp



# ###################################
#  Authentication Client Handles
# ###################################

def getClient():
    '''Get a new instance of the authClient client.

    Parameters
    ----------
    None

    Returns
    -------
    client : authClient
        An authClient object

    Example
    -------
    .. code-block:: python
        new_client = authClient.getClient()
    '''
    return authClient()


ac_client = getClient()


# ##########################################
#  Patch the docstrings for module functions
# ##########################################

login.__doc__ = ac_client.login.__doc__
logout.__doc__ = ac_client.logout.__doc__
passwordReset.__doc__ = ac_client.passwordReset.__doc__

whoAmI.__doc__ = ac_client.whoAmI.__doc__
isAlive.__doc__ = ac_client.isAlive.__doc__
isValidToken.__doc__ = ac_client.isValidToken.__doc__
isValidUser.__doc__ = ac_client.isValidUser.__doc__
isValidPassword.__doc__ = ac_client.isValidPassword.__doc__
hasAccess.__doc__ = ac_client.hasAccess.__doc__
isUserLoggedIn.__doc__ = ac_client.isUserLoggedIn.__doc__
isTokenLoggedIn.__doc__ = ac_client.isTokenLoggedIn.__doc__

set_svc_url.__doc__ = ac_client.set_svc_url.__doc__
get_svc_url.__doc__ = ac_client.get_svc_url.__doc__
set_profile.__doc__ = ac_client.set_profile.__doc__
get_profile.__doc__ = ac_client.get_profile.__doc__


# ####################################################################
#  Py2/Py3 Compatability Utilities
# ####################################################################

def acToString(s):
    '''acToString -- Force a return value to be type 'string' for all
                      Python versions.
    '''
    if is_py3:
        if isinstance(s,bytes):
            strval = str(s.decode())
        elif isinstance(s,str):
            strval = s
    else:
        if isinstance(s,bytes) or isinstance(s,unicode):
            strval = str(s)
        else:
            strval = s

    return strval





import os
try:
    import cPickle as pickle
except:
    import pickle

import BitVector
import logging

logger = logging.getLogger('cache')

class CacheMetaData(object):

    def __init__(self, metaDataFile, blocks, md5sum, size):
        """
        Creates an instance of CacheMetaData for the given file. If the same
        file with the same md5sum already exists in the file cache metadata
        the returned instance is updated with the existing bitmap information
        otherwise a new corresponding bitmap is created
        metaDataFile - name of the metadata file to persist to
        blocks - number of blocks required to store the entire file. If None,
                the file must exist.
        md5sum - md5sum of the file. If None, the file must exist.
        size - Size of the file. If None, the file must exist.
        """

        self.metaDataFile = metaDataFile
        self.blocks = blocks is None and 0 or blocks
        self.md5sum = md5sum
        self.size = size
        self.bitmap = None
        if os.path.exists(self.metaDataFile):
            f = open(self.metaDataFile, 'rU')
            persisted = pickle.load(f)
            if self.md5sum is None or persisted.md5sum == self.md5sum:
                #persisted bitmap still valid. Used that instead
                self.bitmap = persisted.bitmap
                self.size = persisted.size
                self.md5sum = persisted.md5sum
            f.close()

        if self.bitmap is None:
            self.bitmap = BitVector.BitVector(size=self.blocks)

    def __str__(self):
        """To create a print representation that is informative."""
        return "CacheMetaData: metaDataFile=%r bitmap=%r md5sum=%r size=%r" % (self.metaDataFile,
                                                                               self.bitmap,
                                                                               self.md5sum,
                                                                               self.size)

    def __repr__(self):
        return "CacheMetaData(metaDataFile=%r, blocks=%r, md5sum=%r, size=%r)" % (self.metaDataFile,
                                                                                  self.blocks,
                                                                                  self.md5sum,
                                                                                  self.size)
    def setReadBlocks(self, start, end):
        """ To mark several blocks as read (start and end inclusive). """
        startBlock = start
        endBlock = end
        if start < 0:
            startBlock = self.bitmap.length() + start
        if end < 0:
            endBlock = self.bitmap.length() + end
        if startBlock > endBlock:
            raise ValueError('''Incorrect interval, max is %d > %d''' %
                    (startBlock, endBlock))
        for i in range(startBlock, endBlock + 1):
            self.setReadBlock(i)

    def setReadBlock(self, block):
        """ To mark a block as read """
        self.bitmap[block] = 1

    def getBit(self, position):
        """ To return the value of the bit at position """
        return self.bitmap[position]

    def getNumReadBlocks(self):
        """ To return the number of read blocks """
        return self.bitmap.count_bits_sparse()

    def getRange(self, start, end):
        """ To return the range of blocks the client needs to download in order
            to get the given interval of blocks (start and end inclusive)

            Returned result will be the (start, end) tuple where
            start - start of the range
            end - end of the range

            Note: The returned tuple for when all the blocks in the requested
            interval are already downloaded is (None, None)
            """
        startBlock = start
        endBlock = end
        if start < 0:
            startBlock = self.bitmap.length() + start
        if end < 0:
            endBlock = self.bitmap.length() + end
        if startBlock > endBlock:
            raise ValueError('''Incorrect interval''')

        for i in range(startBlock, endBlock + 1):
            if self.bitmap[i] == 0:
                startBlock = i
                break
            if i == endBlock:
                #all the blocks are cached already
                return (None, None)

        for i in reversed(range(startBlock, endBlock + 1)):
            if self.bitmap[i] == 0:
                endBlock = i
                break

        return startBlock, endBlock

    def getNextReadBlock(self, start):
        """
        To return the next block after start that is already read. Returned
        value is the first block that is already read or -1 if none of the
        subsequent blocks is read
        """
        return self.bitmap.next_set_bit(start)

    def delete(self):
        """ To delete bitmap information both from persistence layer
            and from current object """
        os.remove(self.metaDataFile)
        self.bitmap.reset(0)

    def persist(self):
        """To persist cache file metadata for the current file """
        if not os.path.exists(os.path.dirname(self.metaDataFile)):
            os.makedirs(os.path.dirname(self.metaDataFile))
        f = open(self.metaDataFile, 'w+')
        pickle.dump(self, f, -1)
        f.close()

    @staticmethod
    def deleteCacheMetaData(metaDataFile):
        """ To delete an existing cache metadata file """
        os.remove(metaDataFile)



__author__ = 'jjk'

import errno
import os


def mkdir_p(path, mode):
    try:
        os.makedirs(path, mode)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            os.chmod(path, mode)
        else:
            raise OSError(errno.ENOTDIR, "{0} exists and is not a directory".format(path))


#!/usr/bin/env python
#
# QUERYCLIENT -- Client routines for the Data Lab Query Manager Service


from __future__ import print_function

__authors__ = 'Mike Fitzpatrick <fitz@noao.edu>, Matthew Graham <graham@noao.edu>, Data Lab <datalab@noao.edu>'
__version__ = '20190422'  # yyyymmdd


'''
    Client methods for the DataLab Query Manager Service.


Query Manager Client Interface
------------------------------
               isAlive  (svc_url=DEF_SERVICE_URL, timeout=2)
           set_svc_url  (svc_url)
           get_svc_url  ()
           set_profile  (profile)
           get_profile  ()
         list_profiles  (optval, token=None, profile=None, format='text')
         list_profiles  (token=None, profile=None, format='text')

   set_timeout_request  (nsec)
   get_timeout_request  ()
                schema  (value, format='text', profile=None)
                schema  (value='', format='text', profile=None)
              services  (name=None, svc_type=None, format=None,
                         profile='default')

                 query  (token, query, adql=None, sql=None, fmt='csv', out=None,
                         async_=False, profile='default', **kw)
                 query  (optval, adql=None, sql=None, fmt='csv', out=None,
                         async_=False, profile='default', **kw)
                 query  (token=None, adql=None, sql=None, fmt='csv', out=None,
                         async_=False, profile='default', **kw)
                status  (token, jobId)
                status  (optval, jobId=None)
                status  (token=None, jobId=None)
               results  (token, jobId, delete=True)
               results  (optval, jobId=None, delete=True)
               results  (token=None, jobId=None, delete=True)
                 error  (token, jobId)
                 error  (optval, jobId=None)
                 error  (token=None, jobId=None)
                 abort  (token, jobId)
                 abort  (optval, jobId=None)
                 abort  (token=None, jobId=None)

             mydb_list  (optval, table=None)
             mydb_list  (table=None, token=None)
           mydb_create  (token, table, schema, **kw)
           mydb_create  (table, schema, token=None, **kw)
           mydb_insert  (token, table, data, **kw)
           mydb_insert  (table, data, token=None, **kw)
           mydb_import  (token, table, data, **kw)
           mydb_import  (table, data, token=None, **kw)
         mydb_truncate  (token, table)
         mydb_truncate  (table, token=None)
            mydb_index  (token, table, column)
            mydb_index  (table, column, token=None)
             mydb_drop  (token, table)
             mydb_drop  (table, token=None)
           mydb_rename  (token, source, target)
           mydb_rename  (source, target, token=None)
             mydb_copy  (token, source, target)
             mydb_copy  (source, target, token=None)

                  list  (token, table)			# DEPRECATED
                  list  (optval, table=None)		# DEPRECATED
                  list  (table=None, token=None)	# DEPRECATED
                  drop  (token, table)			# DEPRECATED
                  drop  (optval, table=None)		# DEPRECATED
                  drop  (table=None, token=None)	# DEPRECATED


Import via

.. code-block:: python

    from dl import queryClient
'''

import requests
try:
    from urllib import quote_plus               # Python 2
except ImportError:
    from urllib.parse import quote_plus         # Python 3
try:
    from cStringIO import StringIO
except:
    from io import StringIO			# Python 2/3 compatible
import socket
import json
import time
import os
import sys
import collections
import ast, csv
import pandas
from tempfile import NamedTemporaryFile

from dl import storeClient
from dl.helpers.utils import convert
if os.path.isfile('./Util.py'):			# use local dev copy
    from Util import multimethod
    from Util import def_token
else:						# use distribution copy
    from dl.Util import multimethod
    from dl.Util import def_token

is_py3 = sys.version_info.major == 3



# ####################################
#  Query Manager Configuration
# ####################################


# The URL of the QueryManager service to contact.  This may be changed by
# passing a new URL into the set_svc_url() method before beginning.

DEF_SERVICE_ROOT = 'https://datalab.noao.edu'
DAL_SERVICE_URL = 'https://datalab.noao.edu' 	# The base DAL service URL

# Allow the service URL for dev/test systems to override the default.
THIS_HOST = socket.gethostname()			# host name
THIS_IP	  = socket.gethostbyname(socket.gethostname())	# host IP address

if THIS_HOST[:5] == 'dldev':
    DEF_SERVICE_ROOT = 'http://dldev.datalab.noao.edu'
elif THIS_HOST[:6] == 'dltest':
    DEF_SERVICE_ROOT = 'http://dltest.datalab.noao.edu'

DEF_SERVICE_URL = DEF_SERVICE_ROOT + '/query'
SM_SERVICE_URL  = DEF_SERVICE_ROOT + '/storage'

# The requested query 'profile'.  A profile refers to the specific
# machines and services used by the QueryManager on the server.
DEF_PROFILE 	= 'default'

# Use a /tmp/QM_DEBUG file as a way to turn on debugging in the client code.
DEBUG 		= os.path.isfile('/tmp/QM_DEBUG')

# Default sync query timeout default (300sec)
TIMEOUT_REQUEST = 300



# ####################################################################
#  Query Client error class
# ####################################################################

class queryClientError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


# ####################################################################
#  Module Functions
# ####################################################################

# --------------------------------------------------------------------
# ISALIVE -- Ping the Query Manager service to see if it responds.
#
def isAlive(svc_url=DEF_SERVICE_URL, timeout=5):
    return qc_client.isAlive(svc_url=svc_url, timeout=timeout)


# --------------------------------------------------------------------
# SET_SVC_URL -- Set the Query Manager ServiceURL to call.
#
def set_svc_url(svc_url):
    return qc_client.set_svc_url(svc_url)

# --------------------------------------------------------------------
# GET_SVC_URL -- Get the Query Manager ServiceURL being called.
#
def get_svc_url():
    return qc_client.get_svc_url()

# --------------------------------------------------------------------
# SET_PROFILE -- Set the Query Manager service profile to be used.
#
def set_profile(profile):
    return qc_client.set_profile(profile)

# --------------------------------------------------------------------
# GET_PROFILE -- Get the Query Manager service profile being used.
#
def get_profile():
    return qc_client.get_profile()


# -----------------------------
#  Utility Functions
# -----------------------------

# --------------------------------------------------------------------
# LIST_PROFILES -- List the available service profiles.
#
@multimethod('qc',1,False)
def list_profiles(optval, token=None, profile=None, format='text'):
    if optval is not None and len(optval.split('.')) >= 4:
        # optval looks like a token
        return qc_client._list_profiles (token=def_token(optval),
                                         profile=profile, format=format)
    else:
        # optval looks like a profile name
        return qc_client._list_profiles (token=def_token(token), profile=optval,
                                         format=format)

@multimethod('qc',0,False)
def list_profiles(token=None, profile=None, format='text'):
    '''Retrieve the profiles supported by the query manager service.

    Usage:
        list_profiles (token=None, profile=None, format='text')

    MultiMethod Usage:  
    ------------------
            queryClient.list_profiles (token)
            queryClient.list_profiles ()

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    profile : str
        A specific profile configuration to list.  If None, a list of
        profiles available to the given auth token is returned.

    format : str
        Result format: One of 'text' or 'json'

    Returns
    -------
    profiles : list/dict
        A list of the names of the supported profiles or a dictionary of
        the specific profile

    Example
    -------
    .. code-block:: python

        profiles = queryClient.list_profiles()
        profiles = queryClient.list_profiles(token)
    '''
    return qc_client._list_profiles (token=def_token(token), profile=profile,
                                 format=format)

# --------------------------------------------------------------------
# SET_TIMEOUT_REQUEST -- Set the Synchronous query timeout value (in sec).
#
def set_timeout_request(nsec):
    return qc_client.set_timeout_request (nsec)

# --------------------------------------------------------------------
# GET_TIMEOUT_REQUEST -- Get the Synchronous query timeout value (in sec).
#
def get_timeout_request():
    return qc_client.get_timeout_request ()

# --------------------------------------------------------------------
# SCHEMA -- Return information about a data service schema value.
#
@multimethod('qc',1,False)
def schema(value, format='text', profile=None):
    return qc_client._schema (value=value, format=format, profile=profile)

@multimethod('qc',0,False)
def schema(value='', format='text', profile=None):
    '''Return information about a data service schema.

    Usage:
        schema (value='', format='text', profile=None)

    Parameters
    ----------
    value : str
        Schema object to return: Of the form <schema>[.<table>[.<column]]

    profile : str
        The name of the service profile to use. The list of available
        profiles can be retrieved from the service (see function
        :func:`queryClient.list_profiles()`)

    format : str
        Result format:  One of 'text' or 'json'  (NOT CURRENTLY USED)

    Returns
    -------

    Example
    -------
    .. code-block:: python

        # List the available schema
        queryClient.schema("", "text", "default")

        # List the tables in the USNO schema
        queryClient.schema("usno", "text", "default")

        # List the columns of the USNO-A2 table
        queryClient.schema("usno.a2", "text", "default")

        # List the attributes of the USNO-A2 'raj2000' column
        queryClient.schema("usno.a2.raj2000", "text", "default")
    '''
    return qc_client._schema (value=value, format=format, profile=profile)

# --------------------------------------------------------------------
# SERVICES -- List public storage services
#
def services(name=None, svc_type=None, mode='list', profile='default'):
    '''Search or list available data services.

    Usage:
        services (name=None, svc_type=None, mode='list', profile='default')

    Parameters
    ----------
    name : str
        Schema object to return: Of the form <schema>[.<table>[.<column]]

    svc_type : str
        Limit results to specified service type.  Supported options are
	'tap', 'sia', 'scs', or 'vos'.

    mode : str
	Query mode:  

    profile : str
        The name of the service profile to use. The list of available
        profiles can be retrieved from the service (see function
        :func:`queryClient.list_profiles()`)

    Returns
    -------
	If mode is 'list' then a human-readable list of matching services is
	returned.  If mode is 'resolve' then a JSON string of matching services
	is returned un the form "{<svc_name> : <svc_url>, ....}"

    Example
    -------
    .. code-block:: python

        # List the available SIA services
        queryClient.services(svc_type="sia")

        # List the available USNO services, note the '%' matching metacharacter
        queryClient.services(name="usno%")

        # Get the serviceURL of the USNO-A2 table
        queryClient.services(name="usno/a2", mode="resolve")
    '''
    return qc_client.services (name=name, svc_type=svc_type, mode=mode,
                               profile=profile)


# -----------------------------
#  Query Functions
# -----------------------------

# --------------------------------------------------------------------
# QUERY -- Send a query to the Query Manager service
#
@multimethod('qc',2,False)
def query(token, query, adql=None, sql=None, fmt='csv', out=None,
           async_=False, profile='default', **kw):
    return qc_client._query (token=def_token(token), adql=adql, sql=query,
                          fmt=fmt, out=out, async_=async_,
                          profile=profile, **kw)

@multimethod('qc',1,False)
def query(optval, adql=None, sql=None, fmt='csv', out=None, async_=False,
           token=None, profile='default', **kw):
    if optval is not None and optval.lower()[:6] == 'select':
        # optval looks like a query string
        return qc_client._query (token=def_token(None), adql=adql, sql=optval,
                              fmt=fmt, out=out, async_=async_,
                              profile=profile, **kw)
    else:
        # optval is (probably) a token
        return qc_client._query (token=def_token(optval), adql=adql, sql=sql,
                              fmt=fmt, out=out, async_=async_,
                              profile=profile, **kw)

@multimethod('qc',0,False)
def query(token=None, adql=None, sql=None, fmt='csv', out=None, async_=False,
           profile='default', **kw):
    '''Send an SQL or ADQL query to the database or TAP service.

    Usage:
        query (token=None, adql=None, sql=None, fmt='csv', out=None,
               async_=False, profile='default', **kw):

    MultiMethod Usage:
    ------------------
        queryClient.query (token, query, <args>)
        queryClient.query (token | query, <args>)

    Parameters
    ----------
    token : str
        Secure token obtained via :func:`authClient.login()`

    adql : str or None
        ADQL query string that will be passed to the DB query manager, e.g.

        .. code-block:: python

            adql='select top 3 ra,dec from gaia_dr1.gaia_source'

    sql : str or None
        SQL query string that will be passed to the DB query manager, e.g.

        .. code-block:: python

            sql='select ra,dec from gaia_dr1.gaia_source limit 3'

    fmt : str
        Format of result to be returned by the query. Permitted values are:

          * 'csv'          The returned result is a comma-separated string
                           that looks like a csv file (newlines at the end
                           of every row) [DEFAULT]
          * 'ascii'        Same, but the column separator is a tab \t
          * 'array'        Returns a NumPy array
          * 'pandas'       Returns a Pandas DataFrame
          * 'structarray'  Numpy structured array (aka 'record array')
          * 'table'        Returns an Astropy Table object

            The following formats may be used when saving a file to
            virtual storage on the server:

          * 'fits'         FITS binary
          * 'hdf5'         HDF5 file                (NOT YET IMPLEMENTED)
          * 'votable'      An XML-formatted VOTable

    out : str or None
        The output filename to create on the local machine, the URI of a
        VOSpace or MyDB resource to create, or ``None`` if the result is
        to be returned directly to the caller.

    async_ : bool
        If ``True``, the query is Asynchronous, i.e. a job is submitted to
        the DB, and a jobID token is returned the caller. The jobID must
        be then used to check the query's status and to retrieve the result
        (when the job status is ``COMPLETED``) or the error message (when
        the job status is ``ERROR``). Default is ``False``, i.e. the task
        runs a Synchroneous query.

        ``async_`` replaces the previous ``async`` parameter, because ``async``
        was promoted to a keyword in Python 3.7. Users of Python versions
        prior to 3.7 can continue to use the ``async`` keyword.

    profile : str or None
        The Query Manager profile to use for this call.  If ``None`` then
        the default profile is used.  Available profiles may be listed
        using the :func:`queryClient.list_profiles()`

    **kw : dict
        Optional keyword arguments.  Supported keywords currently include:

           wait = False
               Wait for asynchronous queries to complete? If enabled,
               the query() method will submit the job in async mode
               and then poll for results internally before returning.
               the default is to return the job ID immediately and let
               the client poll for job status and return results.

           timeout = 300
               Requested timeout (in seconds) for a query. For a Sync
               query, this value sets a session timeout request in the
               database that will abort the query at the specified time.
               A maximum value of 600 seconds is permitted.  If the
               ``wait`` option is enabled for an ASync query, this is the
               maximum time the query will be allowed to run before an
               abort() is issued on the job.  The maximum timeout for
               an ASync job is 24-hrs (86400 sec).

           poll = 1
               ASync job polling time in seconds.

           verbose = False
               Print verbose messages during ASync job.

    Returns
    -------
    result : str
        If ``async=False``, the return value is the result of the
        query as a formatted string (see ``fmt``). Otherwise the
        result string is a job token, with which later the
        Asynchroneaous query's status can be checked
        (:func:`queryClient.status()`), and the result retrieved (see
        :func:`queryClient.result()`.

    Example
    -------
    Get security token first, see :func:`authClient.login()`. Then:

    .. code-block:: python

        query = 'select ra,dec from gaia_dr1.gaia_source limit 3'
        response = queryClient.query(token, adql=query, fmt='csv')
        print response

    This prints

    .. code::

          ra,dec
          315.002571989537842,35.2662974820284489
          315.00408275885701,35.2665448169895797
          314.996334457679438,35.2673478725552698
    '''
    return qc_client._query (token=def_token(token), adql=adql, sql=sql, 
                             fmt=fmt, out=out, async_=async_, profile=profile,
                             **kw)



# --------------------------------------------------------------------
# STATUS -- Get the status of an Asynchronous query
#
@multimethod('qc',2,False)
def status(token, jobId):
    return qc_client._status (token=def_token(token), jobId=jobId)

@multimethod('qc',1,False)
def status(optval, jobId=None):
    if optval is not None and len(optval.split('.')) >= 4:
        # optval looks like a token
        return qc_client._status (token=def_token(optval), jobId=jobId)
    else:
        # optval is probably a jobId
        return qc_client._status (token=def_token(None), jobId=optval)

@multimethod('qc',0,False)
def status(token=None, jobId=None):
    '''Get the status of an asynchronous query.

    Usage:
        status (token=None, jobId=None)

    MultiMethod Usage:
    ------------------
        queryClient.status (jobId)
        queryClient.status (token, jobId)
        queryClient.status (token, jobId=<id>)
        queryClient.status (jobId=<str>)

    Use the authentication token and the jobId of a previously issued
    asynchronous query to check the query's current status.

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    jobId : str
        The jobId returned when issuing an asynchronous query via
        :func:`queryClient.query()` with ``async=True``.

    Returns
    -------
    status : str

        Either 'QUEUED' or 'EXECUTING' or 'COMPLETED'. If the token &
        jobId combination does not correspond to an actual job, then a
        HTML-formatted error message is returned. If there is a
        problem with the backend, the returned value can be 'ERROR'.

        When status is 'COMPLETED', you can retrieve the results of
        the query via :func:`queryClient.results()`

    Example
    -------
    .. code-block:: python

        import time
        query = 'select ra,dec from gaia_dr1.gaia_source limit 200000'
        jobId = queryClient.query(token, adql=query, fmt='csv', async=True)
        while True:
            status = queryClient.status(token, jobId)
            print "time index =", time.localtime()[5], "   status =", status
            if status == 'COMPLETED':
                break
            time.sleep(1)

    This prints

    .. code::

        time index = 16    status = EXECUTING
        time index = 17    status = EXECUTING
        time index = 18    status = COMPLETED

    '''
    return qc_client._status (token=def_token(token), jobId=jobId)


# --------------------------------------------------------------------
# RESULTS -- Get the results of an Asynchronous query
#
@multimethod('qc',2,False)
def results(token, jobId, delete=True):
    return qc_client._results (token=def_token(token), jobId=jobId, delete=True)

@multimethod('qc',1,False)
def results(optval, jobId=None, delete=True):
    if optval is not None and len(optval.split('.')) >= 4:
        # optval looks like a token
        return qc_client._results (token=def_token(optval), jobId=jobId,
                                delete=delete)
    else:
        # optval is probably a jobId
        return qc_client._results (token=def_token(None), jobId=optval,
                                delete=delete)

@multimethod('qc',0,False)
def results(token=None, jobId=None, delete=True):
    '''Retrieve the results of an asynchronous query, once completed.

    Usage:
        results (token=None, jobId=None, delete=True)

    MultiMethod Usage:
    ------------------
        queryClient.results (jobId)
        queryClient.results (token, jobId)
        queryClient.results (token, jobId=<id>)
        queryClient.results (jobId=<str>)

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    jobId : str
        The jobId returned when issuing an asynchronous query via
        :func:`queryClient.query()` with ``async=True``.

    Returns
    -------
    results : str

    Example
    -------
    .. code-block:: python

        # issue an async query (here a tiny one just for this example)
        query = 'select ra,dec from gaia_dr1.gaia_source limit 3'
        jobId = queryClient.query(token, adql=query, fmt='csv', async=True)

        # ensure job completes...then check status and retrieve results
        time.sleep(4)
        if queryClient.status(token, jobId) == 'COMPLETED':
            results = queryClient.results(token,jobId)
            print type(results)
            print results

    This prints

    .. code::

        <type 'str'>
        ra,dec
        301.37502633933002,44.4946851014515588
        301.371102372343785,44.4953207577355698
        301.385106974224186,44.4963443903961604
    '''
    return qc_client._results (token=def_token(token), jobId=jobId, delete=True)


# --------------------------------------------------------------------
# ERROR -- Get the error message of a failed Asynchronous query
#
@multimethod('qc',2,False)
def error(token, jobId):
    return qc_client._error (token=def_token(token), jobId=jobId)

@multimethod('qc',1,False)
def error(optval, jobId=None):
    if optval is not None and len(optval.split('.')) >= 4:
        # optval looks like a token
        return qc_client._error (token=def_token(optval), jobId=jobId)
    else:
        # optval is probably a jobId
        return qc_client._error (token=def_token(None), jobId=optval)

@multimethod('qc',0,False)
def error(token=None, jobId=None):
    '''Retrieve the error of an asynchronous query, once completed.

    Usage:
        error (token=None, jobId=None)

    MultiMethod Usage:
    ------------------
        queryClient.error (jobId)
        queryClient.error (token, jobId)
        queryClient.error (token, jobId=<id>)
        queryClient.error (jobId=<str>)

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    jobId : str
        The jobId returned when issuing an asynchronous query via
        :func:`queryClient.query()` with ``async=True``.

    Returns
    -------
    error : str

    Example
    -------
    .. code-block:: python

        # issue an async query (here a tiny one just for this example)
        query = 'select ra,dec from gaia_dr1.gaia_source limit 3'
        jobId = queryClient.query(token, adql=query, fmt='csv', async=True)

        # ensure job completes...then check status and retrieve error
        time.sleep(4)
        if queryClient.status(token, jobId) == 'ERROR':
            error = queryClient.error(token,jobId)
            print type(error)
            print error

    This prints

    .. code::

        <type 'str'>
        ra,dec
        301.37502633933002,44.4946851014515588
        301.371102372343785,44.4953207577355698
        301.385106974224186,44.4963443903961604
    '''
    return qc_client._error (token=def_token(token), jobId=jobId)


# --------------------------------------------------------------------
# ABORT -- Abort the specified Asynchronous job.
#
@multimethod('qc',2,False)
def abort(token, jobId):
    return qc_client._abort (token=def_token(token), jobId=jobId)

@multimethod('qc',1,False)
def abort(optval, jobId=None):
    if optval is not None and len(optval.split('.')) >= 4:
        # optval looks like a token
        return qc_client._abort (token=def_token(optval), jobId=jobId)
    else:
        # optval is probably a jobId
        return qc_client._abort (token=def_token(None), jobId=optval)

@multimethod('qc',0,False)
def abort(token=None, jobId=None):
    '''Abort the specified asynchronous job.

    Usage:
        abort (token=None, jobId=None)

    MultiMethod Usage:
    ------------------
        queryClient.abort (token, jobId)
        queryClient.abort (jobId)
        queryClient.abort (token, jobId=<id>)
        queryClient.abort (jobId=<str>)

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    jobId : str
        The jobId to abort.

    Returns
    -------
    results : str

    Example
    -------
    .. code-block:: python

        # issue an async query (here a tiny one just for this example)
        query = 'select ra,dec from gaia_dr1.gaia_source limit 3'
        jobId = queryClient.query(token, adql=query, fmt='csv', async=True)

        # ensure job completes...then check status and retrieve results
        time.sleep(4)
        if queryClient.status(token, jobId) == 'COMPLETED':
            results = queryClient.results(token,jobId)
            print type(results)
            print results

    This prints

    .. code::

        <type 'str'>
        ra,dec
        301.37502633933002,44.4946851014515588
        301.371102372343785,44.4953207577355698
        301.385106974224186,44.4963443903961604
    '''
    return qc_client._abort (token=def_token(token), jobId=jobId)



# -----------------------------
#  MyDB Functions (deprecated)
# -----------------------------

# --------------------------------------------------------------------
# LIST -- List the tables or table schema in a user's MyDB.
#
@multimethod('qc',2,False)
def list(token, table):
    return qc_client.mydb_list (token=def_token(token), table=table)

@multimethod('qc',1,False)
def list(optval, table=None):
    if optval is not None and len(optval.split('.')) >= 4:
        # optval looks like a token
        return qc_client.mydb_list (token=def_token(optval), table=table)
    else:
        # optval is likely a table
        return qc_client.mydb_list (token=def_token(None), table=optval)

@multimethod('qc',0,False)
def list(table=None, token=None):
    '''List the tables or table schema in the user's MyDB.

    Usage:
        list (table=None, token=None)

    MultiMethod Usage:
    ------------------
        queryClient.list (token, table)
        queryClient.list (table)
        queryClient.list (token, table=<id>)
        queryClient.list ()

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    table: str
        The specific table to list (returns the table schema), or
        an empty string to return a list of the names of all tables.

    Returns
    -------
    listing : str
        The list of tables in the user's MyDB or the schema of the
        named table

    Example
    -------
    .. code-block:: python

        # List the tables
        queryClient.list()
    '''
    return qc_client.mydb_list (token=def_token(token), table=table)


# --------------------------------------------------------------------
# DROP -- Drop the named table from a user's MyDB.
#
@multimethod('qc',2,False)
def drop(token, table):
    return qc_client.mydb_drop (token=def_token(token), table=table)

@multimethod('qc',1,False)
def drop(optval, table=None):
    if optval is not None and len(optval.split('.')) >= 4:
        # optval looks like a token
        return qc_client.mydb_drop (token=def_token(optval), table=table)
    else:
        # optval is likely a table
        return qc_client.mydb_drop (token=def_token(None), table=optval)

@multimethod('qc',0,False)
def drop(table=None, token=None):
    '''Drop the specified table from the user's MyDB

    Usage:
        drop (table=None, token=None)

    MultiMethod Usage:
    ------------------
        queryClient.drop (token, table)
        queryClient.drop (table)
        queryClient.drop (token, table=<id>)
        queryClient.drop (table=<str>)

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    table: str
        The specific table to drop

    Returns
    -------

    Example
    -------
    .. code-block:: python

        # List the tables
        queryClient.drop('foo1')
    '''
    return qc_client.mydb_drop (token=def_token(token), table=table)



# -----------------------------
#  MyDB Functions (New API)
# -----------------------------

# --------------------------------------------------------------------
# MYDB_LIST -- List the tables or table schema in a user's MyDB.
#
@multimethod('qc',1,False)
def mydb_list(optval, table=None, index=False):
    if optval is not None and len(optval.split('.')) >= 4:
        # optval looks like a token
        return qc_client._mydb_list (token=def_token(optval), table=table,
                                     index=index)
    else:
        # optval is likely a table name
        return qc_client._mydb_list (token=def_token(None), table=optval,
                                     index=index)

@multimethod('qc',0,False)
def mydb_list(table=None, token=None, index=False):
    '''List the tables or table schema in the user's MyDB.

    Usage:
        mydb_list (table=None, token=None)

    MultiMethod Usage:
    ------------------
        queryClient.mydb_list (table)
        queryClient.mydb_list (token, table=<str>)
        queryClient.mydb_list (table=<str>)

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    table: str
        The specific table to list (returns the table schema), or
        an empty string to return a list of the names of all tables.

    Returns
    -------
    listing : str
        The list of tables in the user's MyDB or the schema of the
        named table

    Example
    -------
    .. code-block:: python

        # List the tables
        queryClient.mydb_list()
    '''
    return qc_client._mydb_list (token=def_token(token), table=table,
                                 index=index)


# --------------------------------------------------------------------
# MYDB_CREATE -- Create a table in the user's MyDB from a local file
# or python data object.
#
@multimethod('qc',3,False)
def mydb_create(token, table, schema, **kw):
    return qc_client._mydb_create (token=def_token(None), table=table,
                              schema=schema, **kw)

@multimethod('qc',2,False)
def mydb_create(table, schema, token=None, **kw):
    '''Create a table in the user's MyDB

    Usage:
        mydb_create (table, schema, token=None, **kw)

    MultiMethod Usage:
    ------------------
        queryClient.mydb_create (token, table, <schema_dict>)
        queryClient.mydb_create (table, <schema_dict>)

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    table: str
        The name of the table to create

    schema: str or dict
        The schema is CSV text containing the name of the column and
        it's PostgreSQL data type.  If set as a 'str' type it is either
        a CSV string, or the name of a file containing the CSV.  If passed
        as a 'dict' type, it is a dictionary object where keys are the
        column names and values are the data types.

    drop: bool   (optional)
        Drop any existing table of the same name before creating new one.

    Returns
    -------

    Example
    -------
    .. code-block:: python

        # Create table in MyDB named 'foo' with columns defined by
        # the file 'schema.txt'.  The schema file contains:
        #
        #     id,text
        #     ra,double precision
        #     dec,double precision
        #
        queryClient.mydb_create ('foo', 'schema.txt')
    '''
    return qc_client._mydb_create (token=def_token(None), table=table,
                              schema=schema, **kw)

#--------------------------------------------------------------------
# MYDB_INSERT -- Insert data into a table in the user's MyDB from a local
# file or python data object.
#
@multimethod('qc',3,False)
def mydb_insert(token, table, data, **kw):
    return qc_client._mydb_insert (token=def_token(token), table=table,
                                   data=data, **kw)

@multimethod('qc',2,False)
def mydb_insert(table, data, token=None, **kw):
    '''Insert data into a table in the user's MyDB

    Usage:
        mydb_insert (table, data, token=None, **kw)

    MultiMethod Usage:
    ------------------
        queryClient.mydb_insert (token, table, <filename>)
        queryClient.mydb_insert (token, table, <data_object>)
        queryClient.mydb_insert (table, <filename>)
        queryClient.mydb_insert (table, <data_object>)

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    table: str
        The name of the table to append

    data: str or data object
        The schema is CSV text containing the name of the column and
        it's PostgreSQL data type.  If set as a 'str' type it is either
        a CSV string, or the name of a file containing the CSV data. If
        passed as a tabular data object, it is converted to CSV and sent
        to the service.

    csv_header: bool        [OPTIONAL]
        If True, then the CSV data object contains a CSV header line, i.e.
        the first line is a row of column names.  Otherwise, no column
        names are assumed and the column order must match the table schema.

    Returns
    -------

    Example
    -------
    .. code-block:: python

        # Insert data into a MyDB table named 'foo'.
        queryClient.mydb_insert ('foo', 'data.csv')
    '''
    return qc_client._mydb_insert (token=def_token(token), table=table,
                                   data=data, **kw)


# --------------------------------------------------------------------
# MYDB_IMPORT -- Import a file or Python object to a MyDB table.
#
@multimethod('qc',3,False)
def mydb_import(token, table, data, **kw):
    try:
        result = qc_client._mydb_import (token=def_token(token), table=table,
                                         data=data, **kw)
    except Exception as e:
        return (str(e))

    return result

@multimethod('qc',2,False)
def mydb_import(table, data, token=None, **kw):
    '''Import data into a table in the user's MyDB

    Usage:
        mydb_import (table, data, token=None, **kw)

    MultiMethod Usage:
    ------------------
        queryClient.mydb_import (token, table, data)
        queryClient.mydb_import (table, data)

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    table: str
        The name of the table to be loaded

    data: str or data object
        The data file or python object to be loaded.  The 'data' value
        may be one of the following types:

            filename	    A CSV file of data
            string		    A string containing CSV data
            Pandas DataFrame    A Pandas DataFrame object
                :                   :        :       :

        Additional object types can be added provided the data can be
        converted to a CSV format.

    schema: str	[OPTIONAL]
        If set, this is a filename or string containing a schema for the
        data table to be created.  A schema contains a comma-delimited row
        for each column containing the column name and it's Postgres data
        type.  If not set, the schema is determined automatically from
        the data.

    drop: bool   	[Optional]
        Drop any existing table of the same name before creating new one.

    verbose: bool   [Optional]
        Be verbose about operations.

    Returns
    -------
        schema	A string containing the table schema
        data_obj	The CSV data to be imported (possibly converted)

    Example
    -------
    .. code-block:: python

        # Import data into a MyDB table named 'foo' from file 'data.csv'.
        schema, data = queryClient.mydb_import ('foo', 'data.csv')
    '''
    try:
        result = qc_client._mydb_import (token=def_token(token), table=table,
                                         data=data, **kw)
    except Exception as e:
        return (str(e))

    return result


# --------------------------------------------------------------------
# MYDB_TRUNCATE -- Truncate a table in the user's MyDB.
#
@multimethod('qc',2,False)
def mydb_truncate(token, table):
    return qc_client._mydb_truncate (token=def_token(token), table=table)

@multimethod('qc',1,False)
def mydb_truncate(table, token=None):
    '''Truncate the specified table in the user's MyDB

    Usage:
        mydb_truncate (table, token=None)

    MultiMethod Usage:
    ------------------
        queryClient.mydb_truncate (token, table)
        queryClient.mydb_truncate (table)
        queryClient.mydb_truncate (token, table=<id>)

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    table: str
        The specific table to drop

    Returns
    -------

    Example
    -------
    .. code-block:: python

        # Truncate the table 'foo'
        queryClient.truncate('foo')
    '''
    return qc_client._mydb_truncate (token=def_token(token), table=table)

# --------------------------------------------------------------------
# MYDB_INDEX -- Index a column in a user's MyDB table.
#
@multimethod('qc',3,False)
def mydb_index(token, table, column, q3c=None, cluster=False, async_=False):
    return qc_client._mydb_index(token=def_token(token), table=table,
                                 column=column, q3c=q3c, cluster=cluster,
                                 async_=async_)

@multimethod('qc',2,False)
def mydb_index(table, column, token=None, q3c=None, cluster=False,
               async_=False):
    return qc_client._mydb_index(token=def_token(token), table=table,
                                 column=column, q3c=q3c, cluster=cluster,
                                 async_=async_)

@multimethod('qc',1,False)
def mydb_index(table, column='', token=None, q3c=None, cluster=False,
               async_=False):
    '''Index the specified column in a table in the user's MyDB

    MultiMethod Usage:
    ------------------
        queryClient.mydb_index (table, colunm)
        queryClient.mydb_index (token, table, column)
        queryClient.mydb_index (table, column, token=None)

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    table: str
        The table to be indexed

    column: str
        The column

    q3c: str
        A comma-delimited list of two column names giving the RA and Dec
        positions (decimal degrees) to be used to Q3C index the table.  If
        None, no Q3C index will be computed.

    cluster: bool
        If enabled, data table will be rewritten to cluster on the Q3C index
        for efficiency.  Only used when 'q3c' columns are specified.

    async_: bool
	If enabled, index commands will be submitted asynchronously.

    Returns
    -------
	command status

    Example
    -------
    .. code-block:: python

	# Index the table's "id" column
        queryClient.index('foo1', 'id') 	

	# Index and cluster the table by position
        queryClient.index('foo1', q3c='ra,dec', cluster=True) 	
    '''
    return qc_client._mydb_index(token=def_token(token), table=table,
                                 column=column, q3c=q3c, cluster=cluster,
                                 async_=async_)

# --------------------------------------------------------------------
# MYDB_DROP -- Drop the named table from a user's MyDB.
#
@multimethod('qc',2,False)
def mydb_drop(token, table):
    return qc_client._mydb_drop (token=def_token(token), table=table)

@multimethod('qc',1,False)
def mydb_drop(table, token=None):
    '''Drop the specified table from the user's MyDB

    Usage:
        mydb_drop (table, token=None)

    MultiMethod Usage:
    ------------------
        queryClient.mydb_drop (token, table)
        queryClient.mydb_drop (table)
        queryClient.mydb_drop (token, table=<id>)

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    table: str
        The specific table to drop

    Returns
    -------

    Example
    -------
    .. code-block:: python

        # Drop the 'foo1' table
        queryClient.drop('foo1')
    '''
    return qc_client._mydb_drop (token=def_token(token), table=table)


# --------------------------------------------------------------------
# MYDB_RENAME -- Rename a table in the user's MyDB.
#
@multimethod('qc',3,False)
def mydb_rename(token, source, target):
    return qc_client._mydb_rename (token=def_token(token),
                                source=source, target=target)

@multimethod('qc',2,False)
def mydb_rename(source, target, token=None):
    '''Rename a table in the user's MyDB to a new name

    Usage:
        mydb_rename (source, target, token=None)

    MultiMethod Usage:
    ------------------
        queryClient.mydb_rename (token, source, target)
        queryClient.mydb_rename (source, target)

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    source: str
        The old table name

    target: str
        The new table name

    Returns
    -------

    Example
    -------
    .. code-block:: python

        # Copy table 'foo' to a new table, 'bar'
        queryClient.mydb_rename ('foo', 'bar')
    '''
    return qc_client._mydb_rename (source, target, token=def_token(token))


# --------------------------------------------------------------------
# MYDB_COPY -- Copy a table in the user's MyDB.
#
@multimethod('qc',3,False)
def mydb_copy(token, source, target):
    return qc_client._mydb_copy (token=def_token(token),
                              source=source, target=target)

@multimethod('qc',2,False)
def mydb_copy(source, target, token=None):
    '''Copy a table in the user's MyDB to a new name

    Usage:
        mydb_copy (source, target, token=None)

    MultiMethod Usage:
    ------------------
        queryClient.mydb_copy (token, source, target)
        queryClient.mydb_copy (source, target)

    Parameters
    ----------
    token : str
        Authentication token (see function :func:`authClient.login()`)

    source: str
        The old table name, i.e. the table to be copied

    target: str
        The new table name, i.e. the table to be created

    Returns
    -------

    Example
    -------
    .. code-block:: python

        # Copy table 'foo' to a new table, 'bar'
        queryClient.mydb_copy ('foo', 'bar')
    '''
    return qc_client._mydb_copy (source, target, token=def_token(token))




# ###################################
#  Query Class procedures
# ###################################

class queryClient (object):
    '''
         QUERYCLIENT -- Client-side methods to access the Data Lab
                        Query Manager Service.
    '''
    def __init__(self, profile=DEF_PROFILE, svc_url=DEF_SERVICE_URL):
        '''Initialize the query client. '''

        self.svc_url = svc_url.strip('/')       # QueryMgr service URL
        self.svc_profile = profile  		# QueryMgr service profile

        self.sm_svc_url = SM_SERVICE_URL        # StorageMgr service URL
        self.hostip = THIS_IP
        self.hostname = THIS_HOST
        self.timeout_request = TIMEOUT_REQUEST
        self.async_wait = False

        # Get the $HOME/.datalab directory.
        self.home = '%s/.datalab' % os.path.expanduser('~')

        self.debug = DEBUG                      # interface debug flag


    def isAlive(self, svc_url=None, timeout=5):
        '''Check whether the QueryManager service at the given URL is
           alive and responding.  This is a simple call to the root
           service URL or ping() method.

        Parameters
        ----------
        svc_url : str
            The Query Service URL to ping.
        timeout : int
            Call will assume to have failed if 'timeout' seconds pass.

        Returns
        -------
        result : bool
            True if service responds properly, False otherwise

        Example
        -------
        .. code-block:: python

            if queryClient.isAlive():
                print ("Query Manager is alive")
        '''
        if svc_url is None:
            svc_url = self.svc_url

        try:
            r = requests.get (svc_url, timeout=timeout)
            resp = r.text
            if r.status_code != 200:
                return False
            elif resp is not None and r.text.lower()[:11] != "hello world":
                return False
        except Exception:
            return False

        return True

    def set_svc_url(self, svc_url):
        '''Set the Query Manager service URL.

        Parameters
        ----------
        svc_url : str
            The service URL of the Query Manager to call.

        Returns
        -------
        Nothing

        Example
        -------
        .. code-block:: python

            queryClient.set_svc_url ("http://localhost:7002")
        '''
        self.svc_url = qcToString(svc_url.strip('/'))

    def get_svc_url(self):
        '''Get the currently-used Query Manager serice URL.

        Parameters
        ----------
        None

        Returns
        -------
        service_url : str
            The currently-used Query Service URL.

        Example
        -------
        .. code-block:: python

            print (queryClient.get_svc_url())
        '''
        return qcToString(self.svc_url)

    def set_profile(self, profile):
        '''Set the service profile to be used.

        Parameters
        ----------
        profile : str
            The name of the profile to use. The list of available profiles
            can be retrieved from the service (see function
            func:`queryClient.list_profiles()`)

        Returns
        -------
        Nothing

        Example
        -------
        .. code-block:: python

            queryClient.set_profile('test')
        '''
        self.svc_profile = qcToString(profile)

    def get_profile(self):
        '''Get the current query profile.

        Parameters
        ----------
        None

        Returns
        -------
        profile : str
            The name of the current profile used with the Query Manager service

        Example
        -------
        .. code-block:: python

            print ("Query Service profile = " + queryClient.get_profile())
        '''
        return qcToString(self.svc_profile)

    def set_timeout_request(self, nsec):
        '''Set the requested Sync query timeout value (in seconds).

        Parameters
        ----------
        nsec : int
            The number of seconds requested before a sync query timeout occurs.
            The service may cap this as a server defined maximum.

        Returns
        -------
        Nothing

        Example
        -------
        .. code-block:: python

            # set the sync query timeout request to 30 seconds
            queryClient.set_timeout_request(30)

        '''
        self.timeout_request = nsec

    def get_timeout_request(self):
        '''Get the current Sync query timeout value.

        Parameters
        ----------
        None

        Returns
        -------
        result: int
            Current sync query timeout value.

        Example
        -------
        .. code-block:: python

            # get the current timeout value
            print (queryClient.get_timeout_request())

        '''
        return self.timeout_request


    # ###########################
    #  Utility Methods
    # ###########################

    @multimethod('_qc',1,True)
    def list_profiles(self, optval, token=None, profile=None, format='text'):
        '''Usage:  queryClient.client.list_profiles (token, ...)
        '''
        if optval is not None and len(optval.split('.')) >= 4:
            # optval looks like a token
            return self._list_profiles (token=def_token(optval),
                                        profile=profile, format=format)
        else:
            # optval looks like a token
            return self._list_profiles (token=def_token(token), profile=optval,
                                        format=format)

    @multimethod('_qc',0,True)
    def list_profiles(self, token=None, profile=None, format='text'):
        '''Usage:  queryClient.client.list_profiles (...)
        '''
        return self._list_profiles(token=def_token(token), profile=profile,
                             format=format)

    def _list_profiles(self, token=None, profile=None, format='text'):
        '''Implementation of the list_profiles() method.
        '''

        headers = self.getHeaders (token)

        dburl = '%s/profiles?' % self.svc_url
        if profile != None and profile != 'None' and profile != '':
            dburl += "profile=%s&" % profile
        dburl += "format=%s" % format

        r = requests.get (dburl, headers=headers)
        profiles = qcToString(r.content)
        if '{' in profiles:
            profiles = json.loads(profiles)

        return qcToString(profiles)


    @multimethod('_qc',1,True)
    def schema(self, value, format='text', profile=None):
        '''Usage:  queryClient.schema ([value])
        '''
        return self._schema (value=value, format=format, profile=profile)

    @multimethod('_qc',0,True)
    def schema(self, value='', format='text', profile=None):
        '''Usage:  queryClient.schema ([value])
        '''
        return self._schema (value=value, format=format, profile=profile)


    def _schema(self, value='', format='text', profile=None, **kw):
        '''Implementation of the schema() method.
        '''
        if profile is None:
           profile = self.svc_profile

        url = '%s/schema?value=%s&format=%s&profile=%s' % \
                (self.svc_url, (value), str(format), str(profile))
        try:
            r = requests.get (url, timeout=2)
            resp = r.text
            return resp
        except Exception:
            raise queryClientError("Error getting schema: " + value)

        return qcToString(resp)


    def services(self, name=None, svc_type=None, mode='list',
                  profile='default'):
        '''Usage:  queryClient.services ()
        '''
        return self._services (name=name, svc_type=svc_type, mode=mode,
                                   profile=profile)

    def _services(self, name=None, svc_type=None, mode='list',
                  profile='default'):
        '''Implementation of the services() method.
        '''
        dburl = '/services?'
        if profile is not None and profile != 'None' and profile != '':
            dburl += ("profile=%s" % profile)
        if name is not None and name != 'None' and name != '':
            dburl += ("&name=%s" % name.replace('%','%25'))
        if svc_type is not None and svc_type != 'None' and svc_type != '':
            dburl += ("&type=%s" % svc_type)
        dburl += "&mode=%s" % mode

        r = self.getFromURL(self.svc_url, dburl, def_token(None))
        svcs = qcToString(r.content)

        return svcs



    # ###########################
    #  Query Methods
    # ###########################

    @multimethod('_qc',2,True)
    def query(self, token, query, adql=None, sql=None, fmt='csv', out=None,
               async_=False, profile='default', **kw):
        '''Usage:  queryClient.query (token)
        '''
        return self._query (token=def_token(token), adql=adql, sql=query,
                            fmt=fmt, out=out, async_=async_,
                            profile=profile, **kw)

    @multimethod('_qc',1,True)
    def query(self, optval, adql=None, sql=None, fmt='csv', out=None,
               async_=False, token=None, profile='default', **kw):
        '''Usage:  queryClient.client.query (token, ...)
        '''
        if optval is not None and optval.lower()[:6] == 'select':
            # optval looks like a query string
            return self._query (token=def_token(None), adql=adql, sql=optval,
                                fmt=fmt, out=out, async_=async_,
                                profile=profile, **kw)
        else:
            # optval is (probably) a token
            return self._query (token=def_token(optval), adql=adql, sql=sql,
                                fmt=fmt, out=out, async_=async_,
                                profile=profile, **kw)

    @multimethod('_qc',0,True)
    def query(self, token=None, adql=None, sql=None, fmt='csv', out=None,
               async_=False, profile='default', **kw):
        '''Usage:  queryClient.client.query (...)
        '''
        return self._query (token=def_token(token), adql=adql, sql=sql,
                            fmt=fmt, out=out, async_=async_,
                            profile=profile, **kw)

    def _query(self, token=None, adql=None, sql=None, fmt='csv', out=None,
              async_=False, profile='default', **kw):
        '''Implementation of the query() method.
        '''

        # Process optional keyword arguments.
        if 'async' in kw:
            async_ = kw['async']

        if 'timeout' in kw: 		# set requested timeout on the query
            timeout = int(kw['timeout'])
        else:
            timeout = self.timeout_request
        self.set_timeout_request (timeout)

        wait = self.async_wait 		# see if we wait for an Async result
        if async_ and 'wait' in kw:
            self.async_wait = wait = kw['wait']

        poll_time = 1 			# see if we wait for an Async result
        if async_ and 'poll' in kw:
            self.async_poll = poll_time = int(kw['poll'])

        verbose = False 		# see if we wait for an Async result
        if async_ and 'poll' in kw:
            verbose = kw['verbose']


        # Set service call headers.
        headers = {'Content-Type': 'text/ascii',
                   'X-DL-TimeoutRequest': str(timeout),
                   'X-DL-ClientVersion': __version__,
                   'X-DL-OriginIP': self.hostip,
                   'X-DL-OriginHost': self.hostname,
                   'X-DL-AuthToken': def_token(token)}  # application/x-sql

        if fmt in ['pandas','array','structarray','table']:
            qfmt = 'csv'
        else:
            qfmt = fmt

        if adql is not None and adql != '':
            # Check for unsupported ADQL functions.
            #if 'q3c_' in adql.lower():
            #    msg = "Q3C functions are not allowed in ADQL queries"
            #    raise queryClientError(msg)
            #if 'healpix_' in adql.lower():
            #    msg = "Healpix functions are is not allowed in ADQL queries"
            #    raise queryClientError(msg)
            #if 'htm_' in adql.lower():
            #    msg = "HTM functions are is not allowed in ADQL queries"
            #    raise queryClientError(msg)

            query = quote_plus(adql)		# URL-encode the query string
            dburl = '%s/query?adql=%s&ofmt=%s&out=%s&async=%s' % (
                self.svc_url, query, qfmt, out, async_)

        elif sql is not None and sql != '':
            query = quote_plus(sql)		# URL-encode the query string
            dburl = '%s/query?sql=%s&ofmt=%s&out=%s&async=%s' % (
                self.svc_url, query, qfmt, out, async_)
        else:
            raise queryClientError("No query specified")

        if profile != "default":        	# append the service profile
            dburl += "&profile=%s" % profile
        else:
            dburl += "&profile=%s" % self.svc_profile

        # Make the service call.
        r = requests.get (dburl, headers=headers)
        if r.status_code != 200:
            raise queryClientError (r.text)
        resp = r.content

        if async_ and wait:
            # Sync query timeouts are handled on the server.  If waiting
            # for an async query, loop until job is completed or the timeout
            # expires.
            jobId = resp
            stat = self._status (token=token, jobId=jobId)
            tval = 0
            while (stat != 'COMPLETED'):
                time.sleep (poll_time)
                try:
                    stat = self._status (token=token, jobId=jobId)
                except Exception as e:
                    raise queryClientError (str(e))
                else:
                    if tval > self.timeout_request:
                        stat = self._abort (token=token, jobId=jobId)
                        break
                    if verbose:
                        tim = tval * poll_time
                        rem = timeout - tim
                        print ('Status = %s; elapsed time: %d, timeout in %d' %
                               (stat, tim, rem))
                tval = tval + poll_time

            if tval > self.timeout_request:
                if verbose:
                    print ('Timeout (%d sec) exceeded' % self.timeout_request)
                raise queryClientError ('Query timeout exceeded')
            elif stat not in ['COMPLETED','ERROR']:
                resp = stat
            elif stat == 'ERROR':
		# Retrieve Async error.
                if verbose:
                    print ('Retrieving error')
                resp = self._error (token=token, jobId=jobId).lower()
            elif stat == 'COMPLETED':
		# Retrieve Async results.  A save to vos/mydb is handled below.
                if verbose:
                    print ('Retrieving results')
                resp = self._results (token=token, jobId=jobId).lower()

        if (out is not None and out != '') and not async_:
            # If we're saving to a local file (e.g. in a notebook directory),
            # the file here.  Results saved to VOSpace or MyDB are handled on
            # the server side.
            if out[:7] == 'file://':
                out = out[7:]
            if ':' not in out or out[:out.index(':')] not in ['vos', 'mydb']:
                file = open (out, 'wb')
                file.write (resp)
                file.close ()
            return 'OK'
        else:
            # Otherwise, simply return the result of the query.
            strval = qcToString(resp)
            if fmt in ['pandas','array','structarray','table']:
                return convert (strval,fmt)
            else:
                return strval


    # --------------------------
    # Async jobs status()
    # --------------------------

    @multimethod('_qc',2,True)
    def status(self, token, jobId):
        '''Usage:  queryClient.status (token, jobID)
        '''
        return self._status (token=def_token(token), jobId=jobId)

    @multimethod('_qc',1,True)
    def status(self, optval, jobId=None):
        '''Usage:  queryClient.status (jobID)
                   queryClient.status (token, jobId=<id>)
        '''
        if optval is not None and len(optval.split('.')) >= 4:
            # optval looks like a token
            return self._status (token=def_token(optval), jobId=jobId)
        else:
            # optval is probably a jobId
            return self._status (token=def_token(None), jobId=optval)

    @multimethod('_qc',0,True)
    def status(self, token=None, jobId=None):
        '''Usage:  queryClient.status (jobID=<str>)
        '''
        return self._status (token=def_token(token), jobId=jobId)

    def _status(self, token=None, jobId=None):
        '''Implementation of the status() method.
        '''
        headers = self.getHeaders (token)

        dburl = '%s/status?jobid=%s' % (self.svc_url, jobId)
        if self.svc_profile != "default":
            dburl += "&profile=%s" % self.svc_profile

        r = requests.get (dburl, headers=headers)
        return qcToString(r.content)


    # --------------------------
    # Async jobs results()
    # --------------------------

    @multimethod('_qc',2,True)
    def results(self, token, jobId, delete=True):
        '''Usage:  queryClient.results (token, jobID)
        '''
        return self._results (token=def_token(token), jobId=jobId,
                              delete=delete)

    @multimethod('_qc',1,True)
    def results(self, optval, jobId=None, delete=True):
        '''Usage:  queryClient.results (jobID)
                   queryClient.results (token, jobId=<id>)
        '''
        if optval is not None and len(optval.split('.')) >= 4:
            # optval looks like a token
            return self._results (token=def_token(optval), jobId=jobId,
                                  delete=delete)
        else:
            # optval is probably a jobId
            return self._results (token=def_token(None), jobId=optval,
                                  delete=delete)

    @multimethod('_qc',0,True)
    def results(self, token=None, jobId=None, delete=True):
        '''Usage:  queryClient.results (jobID=<str>)
        '''
        return self._results (token=def_token(token), jobId=jobId,
                              delete=delete)

    def _results(self, token=None, jobId=None, delete=True):
        '''Implementation of the results() method.
        '''
        headers = self.getHeaders (token)

        dburl = '%s/results?jobid=%s&delete=%s' % (self.svc_url, jobId, delete)
        if self.svc_profile != "default":
            dburl += "&profile=%s" % self.svc_profile

        r = requests.get (dburl, headers=headers)
        return qcToString(r.content)


    # --------------------------
    # Async jobs error()
    # --------------------------

    @multimethod('_qc',2,True)
    def error(self, token, jobId):
        '''Usage:  queryClient.error (token, jobID)
        '''
        return self._error (token=def_token(token), jobId=jobId)

    @multimethod('_qc',1,True)
    def error(self, optval, jobId=None):
        '''Usage:  queryClient.error (jobID)
                     queryClient.error (token, jobId=<id>)
        '''
        if optval is not None and len(optval.split('.')) >= 4:
            # optval looks like a token
            return self._error (token=def_token(optval), jobId=jobId)
        else:
            # optval is probably a jobId
            return self._error (token=def_token(None), jobId=optval)

    @multimethod('_qc',0,True)
    def error(self, token=None, jobId=None):
        '''Usage:  queryClient.error (jobID=<str>)
        '''
        return self._error (token=def_token(token), jobId=jobId)

    def _error(self, token=None, jobId=None):
        '''Implementation of the error() method.
        '''
        headers = self.getHeaders (token)

        dburl = '%s/error?jobid=%s' % (self.svc_url, jobId)
        if self.svc_profile != "default":
            dburl += "&profile=%s" % self.svc_profile

        r = requests.get (dburl, headers=headers)
        return qcToString(r.content)


    # --------------------------
    # Async jobs abort()
    # --------------------------

    @multimethod('_qc',2,True)
    def abort(self, token, jobId):
        '''Usage:  queryClient.abort (token, jobID)
        '''
        return self._abort (token=def_token(token), jobId=jobId)

    @multimethod('_qc',1,True)
    def abort(self, optval, jobId=None):
        '''Usage:  queryClient.abort (jobID)
                   queryClient.abort (token, jobId=<id>)
        '''
        if optval is not None and len(optval.split('.')) >= 4:
            # optval looks like a token
            return self._abort (token=def_token(optval), jobId=jobId)
        else:
            # optval is probably a jobId
            return self._abort (token=def_token(None), jobId=optval)

    @multimethod('_qc',0,True)
    def abort(self, token=None, jobId=None):
        '''Usage:  queryClient.abort (jobID=<str>)
        '''
        return self._abort (token=def_token(token), jobId=jobId)

    def _abort(self, token=None, jobId=None):
        '''Implementation of the abort() method.
        '''
        headers = self.getHeaders (token)

        dburl = '%s/abort?jobid=%s' % (self.svc_url, jobId)
        if self.svc_profile != "default":
            dburl += "&profile=%s" % self.svc_profile

        r = requests.get (dburl, headers=headers)
        return qcToString(r.content)


#=========================================================================

    # ###########################
    #  MyDB Methods
    # ###########################

    # LIST -- List the tables or table schema in the user's MyDB.
    #
    @multimethod('_qc',2,True)
    def list(self, token, table):
        '''Usage:  queryClient.list (token, table)
        '''
        return self.mydb_list (token=def_token(token), table=table)

    @multimethod('_qc',1,True)
    def list(self, optval, table=None):
        '''Usage:  queryClient.list (table)
                     queryClient.list (token, table=<id>)
        '''
        if optval is not None and len(optval.split('.')) >= 4:
            # optval looks like a token
            return self.mydb_list (token=def_token(optval), table=table)
        else:
            # optval is probably a table
            return self.mydb_list (token=def_token(None), table=optval)

    @multimethod('_qc',0,True)
    def list(self, token=None, table=None):
        '''Usage:  queryClient.list (table=<str>)
        '''
        return self.mydb_list (token=def_token(token), table=table)


    # DROP -- Drop the specified table from the user's MyDB
    #
    @multimethod('_qc',2,True)
    def drop(self, token, table):
        '''Usage:  queryClient.drop (token, table)
        '''
        return self.mydb_drop (token=def_token(token), table=table)

    @multimethod('_qc',1,True)
    def drop(self, optval, table=None):
        '''Usage:  queryClient.drop (table)
                     queryClient.drop (token, table=<id>)
        '''
        if optval is not None and len(optval.split('.')) >= 4:
            # optval looks like a token
            return self.mydb_drop (token=def_token(optval), table=table)
        else:
            # optval is probably a table
            return self._drop (token=def_token(None), table=optval)

    @multimethod('_qc',0,True)
    def drop(self, token=None, table=None):
        '''Usage:  queryClient.drop (table=<str>)
        '''
        return self.mydb_drop (token=def_token(token), table=table)


    # -----------------------------
    #  MyDB Functions (New API)
    # -----------------------------

    # --------------------------------------------------------------------
    # MYDB_LIST -- List the tables or table schema in a user's MyDB.
    #
    @multimethod('_qc',1,True)
    def mydb_list(self, optval, table=None, index=False):
        '''Usage:  queryClient.mydb_list (table)
                     queryClient.mydb_list (token, table=<str>)
        '''
        if optval is not None and len(optval.split('.')) >= 4:
            # optval looks like a token
            return self._mydb_list (token=def_token(optval), table=table,
                                    index=index)
        else:
            # optval is probably a table
            return self._mydb_list (token=def_token(None), table=optval,
                                    index=index)

    @multimethod('_qc',0,True)
    def mydb_list(self, token=None, table=None, index=False):
        '''Usage:  queryClient.mydb_list (table=<str>)
        '''
        return self._mydb_list (token=def_token(token), table=table,
                                index=index)

    def _mydb_list(self, token=None, table=None, index=False):
        '''Implementation of the mydb_list() method.
        '''
        headers = self.getHeaders (token)

        if table is None:
            table = ''
        dburl = '%s/list?table=%s&index=%s' % (self.svc_url, table, str(index))
        if self.svc_profile != "default":
            dburl += "&profile=%s" % self.svc_profile

        r = requests.get (dburl, headers=headers)
        return qcToString(r.content)


    # --------------------------------------------------------------------
    # MYDB_CREATE -- Copy a table in the user's MyDB.
    #
    @multimethod('_qc',3,True)
    def mydb_create(self, token, table, schema, **kw):
        '''Usage:  queryClient.mydb_create (token, table, <schema_dict>)
        '''
        return self._mydb_create (token=def_token(None), table=table,
                                  schema=schema, **kw)

    @multimethod('_qc',2,True)
    def mydb_create(self, table, schema, token=None, **kw):
        '''Usage:  queryClient.mydb_create (table, <schema_dict>)
        '''
        return self._mydb_create (token=def_token(None), table=table,
                                  schema=schema, **kw)

    def _mydb_create(self, token, table, schema, **kw):
        '''Implementation of the mydb_create() method.
        '''
        # Set the request headers.
        headers = self.getHeaders (token)
        headers['Content'] = 'text/ascii'

        params = { 'table' : table}
        dburl = '%s/create' % (self.svc_url)

        drop = True			# drop table if exists
        verbose = False
        verbose = False			# verbose output
        if 'verbose' in kw:
            verbose = kw['verbose']
        if 'drop' in kw:
            drop = kw['drop']
        params['verbose'] = str(verbose)
        params['drop'] = str(drop)
        params['profile'] = self.svc_profile

        # Schema can be a dictionary, a CSV string, or the name of a file.
        if isinstance (schema,str):
            if os.path.exists (schema):
                with open(schema, 'r') as f:
                    s = f.read()
                params['schema'] = s
            else:
                params['schema'] = schema

        elif isinstance (schema,collections.OrderedDict):
            # We can't use a regular 'dict' object because the key ordered isn't
            # guaranteed to be preserved, but allow OrderedDict.
            s = ''
            for i in schema:
               s += i + ',' + schema[i] + '\n'
            params['schema'] = s

        r = requests.post (dburl, params=params, headers=headers)

        if r.content[:5].lower() == 'error':
            raise queryClientError (qcToString(r.content))
        else:
            return 'OK'


    # --------------------------------------------------------------------
    # MYDB_INSERT -- Insert data into a table in the user's MyDB from a local
    # file or python data object.
    #
    @multimethod('_qc',3,True)
    def mydb_insert(self, token, table, data, **kw):
        '''Usage:  queryClient.mydb_insert (token, table, <filename>)
                   queryClient.mydb_insert (token, table, <data_object>)
        '''
        return self._mydb_insert (token=def_token(token), table=table,
                                  data=data, **kw)

    @multimethod('_qc',2,True)
    def mydb_insert(self, table, data, token=None, **kw):
        '''Usage:  queryClient.mydb_insert (table, <filename>)
                   queryClient.mydb_insert (table, <data_object>)
        '''
        return self._mydb_insert (token=def_token(token), table=table,
                                  data=data, **kw)

    def _mydb_insert(self, token=None, table=None, data=None, **kw):
        '''Implementation of the mydb_create() method.
        '''
        # Get optional parameters.
        csv_header = (kw['csv_header'] if 'csv_header' in kw else True)
        verbose = (kw['drop'] if 'drop' in kw else False)

        # Set up the request headers and initialize.
        headers = self.getHeaders (token)
        params['profile'] = self.svc_profile
        dburl = '%s/ingest' % (self.svc_url)

        # Data can be the name of a CSV file or a python tablular object that
        # can be converted.
        tmp_file = NamedTemporaryFile(delete=True, dir='/tmp').name
        if isinstance (data, str):
            params = { 'table' : table,
                       'csv_header' : str(csv_header) }
            if data.startswith ('http://') or data.startswith('https://') or \
               data.startswith ('vos://'):
                    # Passing a URI in the filename to be loaded on server-side
                    params['filename'] = data

            elif os.path.exists (data):
                # Upload the file to the staging area.
                data_name = os.path.basename (data)
                storeClient.chunked_upload (token, data, data_name)
                params['filename'] = data_name

            else:
                # Upload the file to the staging area.
                with open(tmp_file, 'w') as f:
                    f.write(data)
                f.close()
                tmp_name = os.path.basename (tmp_file)
                storeClient.chunked_upload(token, tmp_file, tmp_name)
                params['filename'] = tmp_name

        elif isinstance (data, pandas.core.frame.DataFrame):
            # Convert a DataFrame to a CSV string object.
            schema, data_to_load = self.getSchema (data)

            with open (tmp_file, 'w') as f:
                f.write(data_to_load)
            f.close()
            tmp_name = os.path.basename (tmp_file)
            storeClient.chunked_upload(token, tmp_file, tmp_name)
            params['filename'] = tmp_name

        else:
            pass

        r = requests.post (dburl, params=params, headers=headers)

        if tmp_file is not None and os.path.exists(tmp_file):
            os.remove (tmp_file)
        if verbose or self.debug:
            print (str(r.text))

        if r.content[:5].lower() == 'error':
            raise queryClientError (qcToString(r.content))
        else:
            return 'OK'


    # --------------------------------------------------------------------
    # MYDB_IMPORT -- Import a file or Python object to a MyDB table.
    #
    @multimethod('_qc',3,True)
    def mydb_import(self, token, table, data, **kw):
        '''Usage:  queryClient.mydb_import (token, table, data)
        '''
        return self._mydb_import (token=def_token(token), table=table,
                                  data=data, **kw)

    @multimethod('_qc',2,True)
    def mydb_import(self, table, data, token=None, **kw):
        '''Usage:  queryClient.mydb_import (table, data)
        '''
        return self._mydb_import (token=def_token(token), table=table,
                                  data=data, **kw)

    def _mydb_import(self, token=None, table=None, data=None, **kw):
        '''Implementation of the mydb_create() method.
        '''
        # Get optional parameters.
        csv_header = (kw['csv_header'] if 'csv_header' in kw else True)
        verbose = (kw['drop'] if 'drop' in kw else False)
        drop = (kw['drop'] if 'drop' in kw else True)
        delimiter = (kw['delimiter'] if 'delimiter' in kw else ',')

        # Set up the request headers and initialize.
        headers = self.getHeaders (token)
        dburl = '%s/import' % (self.svc_url)
        if self.svc_profile != "default":
            dburl += "&profile=%s" % self.svc_profile

        # Data can be the name of a CSV file or a python tablular object that
        # can be converted.
        tmp_file = NamedTemporaryFile(delete=True, dir='/tmp').name
        params = { 'table' : table,
                   'delimiter' : str(delimiter),
                   'drop' : str(drop),
                   'profile' : self.svc_profile,
                   'csv_header' : str(csv_header) }

        if isinstance (data, str):
            if data.startswith ('http://') or \
               data.startswith ('https://') or \
               data.startswith ('vos://'):
                    # Passing a URI in the filename to be loaded on server-side
                    params['filename'] = data

            elif os.path.exists (data):
                # Upload the file to the staging area.
                data_name = os.path.basename (data)
                storeClient.chunked_upload (token, data, data_name)
                params['filename'] = data_name

            else:
                # Upload the CSV string to the staging area.
                with open (tmp_file, 'w') as f:
                    f.write(data)
                f.close()
                tmp_name = os.path.basename (tmp_file)
                storeClient.chunked_upload(token, tmp_file, tmp_name)
                params['filename'] = tmp_name

        elif isinstance (data, pandas.core.frame.DataFrame):
            # Convert a DataFrame to a CSV string object.
            schema, data_to_load = self.getSchema (data)

            with open (tmp_file, 'w') as f:
                f.write(data_to_load)
            f.close()
            tmp_name = os.path.basename (tmp_file)
            storeClient.chunked_upload(token, tmp_file, tmp_name)
            params['filename'] = tmp_name

        else:
            # Reserved for future format support.
            pass

        # Execute the service call.
        r = requests.post (dburl, params=params, headers=headers)

        if verbose or self.debug:
            print (qcToString (r.content))
        if tmp_file is not None and os.path.exists(tmp_file):
            os.remove(tmp_file)

        if r.content[:5].lower() == 'error' or r.status_code != 200:
            raise queryClientError (qcToString(r.content))
        else:
            return 'OK'


    # --------------------------------------------------------------------
    # MYDB_TRUNCATE -- Truncate a table in the user's MyDB.
    #
    @multimethod('_qc',2,True)
    def mydb_truncate(self, token, table):
        '''Usage:  queryClient.mydb_truncate (token, table)
        '''
        return self._mydb_truncate (token=def_token(token), table=table)

    @multimethod('_qc',1,True)
    def mydb_truncate(self, table, token=None):
        '''Usage:  queryClient.mydb_truncate (table)
        '''
        return self._mydb_truncate (token=def_token(token), table=table)

    def _mydb_truncate(self, token=None, table=None):
        '''Implementation of the mydb_truncate() method.
        '''
        headers = self.getHeaders (token)

        dburl = '%s/truncate?table=%s' % (self.svc_url, table)
        if self.svc_profile != "default":
            dburl += "&profile=%s" % self.svc_profile

        r = requests.get (dburl, headers=headers)
        if r.content[:5].lower() == 'error':
            return qcToString(r.content)
        else:
            return 'OK'



    # --------------------------------------------------------------------
    # MYDB_INDEX -- Index a column in a user's MyDB table.
    #
    @multimethod('_qc',3,True)
    def mydb_index(self, token, table, column, q3c=None, cluster=False,
                    async_=False):
        '''Usage:  queryClient.mydb_index (token, table, column)
        '''
        return self._mydb_index(token=def_token(token), table=table,
                                column=column, q3c=q3c, cluster=cluster,
                                async_=async_)

    @multimethod('_qc',2,True)
    def mydb_index(self, table, column, token=None, q3c=None, cluster=False,
                    async_=False):
        '''Usage:  queryClient.mydb_index (table, colunm)
        '''
        return self._mydb_index(token=def_token(token), table=table,
                                column=column, q3c=q3c, cluster=cluster,
                                async_=async_)

    def _mydb_index(self, token=None, table=None, column='', q3c=None,
                    cluster=False, async_=False):
        '''Implementation of the mydb_index() method.
        '''
        headers = self.getHeaders (token)

        if async_:
            def async_call (url, params, headers):
                r = requests.get (url, params=params, headers=headers)
                return qcToString(r.content)

            from multiprocessing.pool import ThreadPool
            params = { 'tbl' : table,
                       'col' : column,
                       'profile' : self.svc_profile }
            if q3c is not None:
                params['q3c'] = q3c
                params['cluster'] = cluster

            pool = ThreadPool(processes=1)
            dburl = '%s/index' % self.svc_url
            res = pool.apply_async (async_call, (dburl, params, headers))
            return 'OK'
        else:
            dburl = '%s/index?tbl=%s&col=%s' % (self.svc_url, table, column)
            if q3c is not None:
                dburl += '&q3c=%s&cluster=%s' % (q3c, cluster)
            if self.svc_profile != "default":
                dburl += "&profile=%s" % self.svc_profile

            r = requests.get (dburl, headers=headers)
            if r.content[:5].lower() == 'error':
                return qcToString(r.content)
            else:
                return 'OK'



    # --------------------------------------------------------------------
    # MYDB_DROP -- Drop the named table from a user's MyDB.
    #
    @multimethod('_qc',2,True)
    def mydb_drop(self, token, table):
        '''Usage:  queryClient.mydb_drop (token, table)
        '''
        return self._mydb_drop (token=def_token(token), table=table)

    @multimethod('_qc',1,True)
    def mydb_drop(self, table, token=None):
        '''Usage:  queryClient.mydb_drop (table)
                     queryClient.mydb_drop (token, table=<id>)
        '''
        return self._mydb_drop (token=def_token(token), table=table)

    def _mydb_drop(self, token=None, table=None):
        '''Implementation of the mydb_drop() method.
        '''
        headers = self.getHeaders (token)

        dburl = '%s/delete?table=%s' % (self.svc_url, table)
        if self.svc_profile != "default":
            dburl += "&profile=%s" % self.svc_profile

        r = requests.get (dburl, headers=headers)
        if r.content[:5].lower() == 'error':
            return qcToString(r.content)
        else:
            return 'OK'


    # --------------------------------------------------------------------
    # MYDB_RENAME -- Rename a table in the user's MyDB.
    #
    @multimethod('_qc',3,True)
    def mydb_rename(self, token, source, target):
        '''Usage:  queryClient.mydb_rename (token, source, target)
        '''
        return self._mydb_rename (token=def_token(token),
                                    source=source, target=target)

    @multimethod('_qc',2,True)
    def mydb_rename(self, source, target, token=None):
        '''Usage:  queryClient.mydb_rename (source, target)
        '''
        return self._mydb_rename (source=source, target=target,
                                  token=def_token(token))

    def _mydb_rename(self, source='', target='', token=None):
        '''Implementation of the mydb_rename() method.
        '''
        headers = self.getHeaders (token)

        dburl = '%s/rename?source=%s&target=%s' % (self.svc_url, source, target)
        if self.svc_profile != "default":
            dburl += "&profile=%s" % self.svc_profile

        r = requests.get (dburl, headers=headers)
        if r.content[:5].lower() == 'error':
            return qcToString(r.content)
        else:
            return 'OK'

    # --------------------------------------------------------------------
    # MYDB_COPY -- Copy a table in the user's MyDB.
    #
    @multimethod('_qc',3,True)
    def mydb_copy(self, token, source, target):
        '''Usage:  queryClient.mydb_copy (token, source, target)
        '''
        return self._mydb_copy (token=def_token(token),
                                  source=source, target=target)

    @multimethod('_qc',2,True)
    def mydb_copy(self, source, target, token=None):
        '''Usage:  queryClient.mydb_copy (source, target)
        '''
        return self._mydb_copy (source=source, target=target,
                                token=def_token(token))

    def _mydb_copy(self, source='', target='', token=None):
        '''Implementation of the mydb_copy() method.
        '''
        headers = self.getHeaders (token)

        dburl = '%s/copy?source=%s&target=%s' % (self.svc_url, source, target)
        if self.svc_profile != "default":
            dburl += "&profile=%s" % self.svc_profile

        r = requests.get (dburl, headers=headers)

        if r.content[:5].lower() == 'error':
            return qcToString(r.content)
        else:
            return 'OK'


    @staticmethod
    def pretty_print_POST(req):
        '''
        At this point it is completely built and ready
        to be fired; it is "prepared".

        However pay attention at the formatting used in
        this function because it is programmed to be pretty
        printed and may differ from the actual request.
        '''
        print('{}\n{}\n{}\n\n{}'.format(
            '-----------START-----------',
            req.method + ' ' + req.url,
            '\n'.join('{}: {}'.format(k, v) for k, v in req.headers.items()),
            req.body,
        ))


    # ############################################
    #    DEPRECATED / NOT-YET-IMPLEMENTED
    # ############################################

    # SIAQUERY -- Send a SIA query to the query manager service
    #
    def siaquery(self, token, input=None, out=None, search=0.5):
        '''Send a SIA (Simple Image Access) query to the query manager service
        '''

        headers = self.getHeaders (token)
        user, uid, gid, hash = token.strip().split('.', 3)

        shortname = '%s_%s' % (uid, input[input.rfind('/') + 1:])
        if input[:input.find(':')] not in ['vos', 'mydb']:
            # Need to set this from config?
            target = 'vos://datalab.noao.edu!vospace/siawork/%s' % shortname
            r = requests.get (SM_SERVICE_URL + "/put?name=%s" %
                             target, headers=headers)
            file = open(input).read()

            headers2 = {'Content-type': 'application/octet-stream',
                        'X-DL-AuthToken': token}
            requests.put(r.content, data=file, headers=headers2)

        dburl = '%s/sia?in=%s&radius=%s&out=%s' % (
            self.svc_url, shortname, search, out)
        r = requests.get (dburl, headers=headers)

        if out is not None:
            if out[:out.index(':')] not in ['vos', 'mydb']:
                file = open(out, 'wb')
                file.write(r.content)
                file.close()
        else:
            return qcToString(r.content)


    # CONEQUERY -- Send a cone search query to the query manager service
    #
    def conequery(self, token, input=None, out=None, schema=None, table=None, ra=None, dec=None, search=0.5):
        '''Send a cone search query to the consearch service
        '''

        headers = self.getHeaders (token)
        user, uid, gid, hash = token.strip().split('.', 3)

    #    shortname = '%s_%s' % (uid, input[input.rfind('/') + 1:])
    #    if input[:input.find(':')] not in ['vos', 'mydb']:
    #        # Need to set this from config?
    #        target = 'vos://datalab.noao.edu!vospace/siawork/%s' % shortname
    #        r = requests.get (SM_SERVICE_URL + "/put?name=%s" %
    #                         target, headers={'X-DL-AuthToken': token})
    #        file = open(input).read()
    #
    #        headers2 = {'Content-type': 'application/octet-stream',
    #                    'X-DL-AuthToken': token}
    #        requests.put(r.content, data=file, headers=headers2)

        dburl = '%s/scs/%s/%s?ra=%s&dec=%s&radius=%s' % (
            DAL_SERVICE_URL, schema, table, ra, dec)
        r = requests.get (dburl, headers=headers)

        if out is not None:
            if out[:out.index(':')] not in ['vos', 'mydb']:
                file = open(out, 'wb')
                file.write(r.content)
                file.close()
        else:
            return qcToString(r.content)


    # -------------------------------------------------------
    #  Private Utility Methods
    # -------------------------------------------------------

    def getHeaders(self, token):
        '''Get default tracking headers,
        '''
        tok = def_token(token)
        user, uid, gid, hash = tok.strip().split('.', 3)
        hdrs = {#'Content-Type': 'text/ascii',
                'X-DL-ClientVersion': __version__,
                'X-DL-OriginIP': self.hostip,
                'X-DL-OriginHost': self.hostname,
                'X-DL-User': user,
                'X-DL-AuthToken': tok}                  # application/x-sql

        return hdrs

    def getFromURL(self, svc_url, path, token):
        '''Get something from a URL.  Return a 'response' object
        '''
        try:
            hdrs = self.getHeaders (token)
            resp = requests.get("%s%s" % (svc_url, path), headers=hdrs)

        except Exception as e:
            raise queryClientError(str(e))
        return resp

    def chunked_upload(self, token, local_file, remote_file):
        '''A streaming file uploader.
        '''
        debug = False
        init = True
        CHUNK_SIZE = 4 * 1024 * 1024                   # 16MB chunks
        url = '%s/xfer' % (self.svc_url)

        # Get the size of the file to be transferred.
        fsize = os.stat(local_file).st_size
        nchunks = fsize / CHUNK_SIZE + 1
        if (debug): print ('Upload in %d chunks ....' % nchunks)
        with open(local_file, 'rb') as f:
            try:
                nsent = 0
                while nsent < fsize:
                    data = f.read(CHUNK_SIZE)
                    requests.post (url, data,
                        headers={'Content-type': 'application/octet-stream',
                                 'X-DL-FileName': remote_file,
                                 'X-DL-InitXfer': str(init),
                                 'X-DL-AuthToken': token})
                    nsent += len(data)
                    if init: init = False
            except Exception as e:
                raise queryClientError ('Upload error: ' + str(e))

    def dataType(self, val, current_type):
        '''Lexically scan a value to determine the datatype.
        '''
        try:
            # Evaluates numbers to an appropriate type, and strings an error
            t = ast.literal_eval(val.strip())
        except ValueError:
            return 'text'
        except SyntaxError:
            return 'text'
        if type(t) in [int, float]:
           if (type(t) in [int]) and current_type not in ['float', 'varchar']:
               # Use smallest possible int type
               if (-32768 < t < 32767) and current_type not in ['int', 'bigint']:
                   return 'smallint'
               elif (-2147483648 < t < 2147483647) and current_type not in ['bigint']:
                   return 'int'
               else:
                   return 'bigint'
           if type(t) is float and current_type not in ['varchar']:
               return ('float' if len(val) < 6 else 'double precision')
        else:
            return 'text'

    def getSchema(self, data, **kw):
        '''Generate a schema for mydb_create() from a CSV file or data object.
        '''
        data_obj = data
        if isinstance (data, str):
            # If a file, assume it is CSV but allow the 'delimiter' et al opts
            if os.path.exists (data):
                reader = csv.reader(open(data, 'r'), **kw)
            else:
                reader = csv.reader(StringIO(data), **kw)
        elif isinstance (data, pandas.core.frame.DataFrame):
            data_obj = data.to_csv(index=False)
            reader = csv.reader(StringIO(data_obj), **kw)
        else:
            print ('Unsupported data format')
            return '', data

        # TODO: Check behavior when no CSV header and data row contains floats

        longest, headers, type_list = [], [], []
        nrows = 10
        for row in reader:
            if len(headers) == 0:		# First row of CSV
                headers = row
                for col in row:
                    longest.append(0)
                    type_list.append('')
            else:
                for i in range(len(row)):
                    if type_list[i] == 'varchar' or row[i] == 'NA':
                        # NA is the csv null value
                        pass
                    else:
                        var_type = self.dataType(row[i], type_list[i])
                        type_list[i] = var_type
                if len(row[i]) > longest[i]:
                    longest[i] = len(row[i])
            if nrows == 0:			# Only read first 10 rows
                break
            nrows = nrows - 1

        schema = ''
        for i in range(len(headers)):
            if type_list[i] == 'text':
                schema += '{},text\n'.format(headers[i].lower().replace('.','_'))
            else:
                schema += '{},{}\n'.format(headers[i].lower().replace('.','_'), type_list[i])

        return schema, data_obj



# ###################################
#  Query Client Handles
# ###################################

def getClient(profile=DEF_PROFILE, svc_url=DEF_SERVICE_URL):
    '''Create a new queryClient object and set a default profile.
    '''
    return queryClient(profile=profile, svc_url=svc_url)

# The default client handle for the module.
qc_client = getClient(profile=DEF_PROFILE, svc_url=DEF_SERVICE_URL)



# ##########################################
#  Patch the docstrings for module functions
#  that aren't MultiMethods.
# ##########################################

isAlive.__doc__ = qc_client.isAlive.__doc__
services.__doc__ = qc_client.services.__doc__
set_svc_url.__doc__ = qc_client.set_svc_url.__doc__
get_svc_url.__doc__ = qc_client.get_svc_url.__doc__
set_profile.__doc__ = qc_client.set_profile.__doc__
get_profile.__doc__ = qc_client.get_profile.__doc__
set_timeout_request.__doc__ = qc_client.set_timeout_request.__doc__
get_timeout_request.__doc__ = qc_client.get_timeout_request.__doc__


# ####################################################################
#  Py2/Py3 Compatability Utilities
# ####################################################################

def qcToString(s):
    '''qcToString -- Force a return value to be type 'string' for all
                      Python versions.
    '''
    strval = s
    if is_py3 and isinstance(s,bytes):
        strval = str(s.decode())
    elif not is_py3 and (isinstance(s,bytes) or isinstance(s,unicode)):
        strval = str(s)
    else:
        strval = s

    return strval
